from typing import Optional, Sequence

import numpy as np


class GraphData:
    """GraphData class

    This data class is almost same as `torch_geometric.data.Data
    <https://pytorch-geometric.readthedocs.io/en/latest/modules/data.html#torch_geometric.data.Data>`_.

    Attributes
    ----------
    node_features: np.ndarray
        Node feature matrix with shape [num_nodes, num_node_features]
    edge_index: np.ndarray, dtype int
        Graph connectivity in COO format with shape [2, num_edges]
    edge_features: np.ndarray, optional (default None)
        Edge feature matrix with shape [num_edges, num_edge_features]
    node_pos_features: np.ndarray, optional (default None)
        Node position matrix with shape [num_nodes, num_dimensions].
    num_nodes: int
        The number of nodes in the graph
    num_node_features: int
        The number of features per node in the graph
    num_edges: int
        The number of edges in the graph
    num_edges_features: int, optional (default None)
        The number of features per edge in the graph

    Examples
    --------
    >>> import numpy as np
    >>> node_features = np.random.rand(5, 10)
    >>> edge_index = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]], dtype=np.int64)
    >>> edge_features = np.random.rand(5, 5)
    >>> global_features = np.random.random(5)
    >>> graph = GraphData(node_features, edge_index, edge_features, z=global_features)
    >>> graph
    GraphData(node_features=[5, 10], edge_index=[2, 5], edge_features=[5, 5], z=[5])
    """

    def __init__(self,
                 node_features: np.ndarray,
                 edge_index: np.ndarray,
                 edge_features: Optional[np.ndarray] = None,
                 node_pos_features: Optional[np.ndarray] = None,
                 **kwargs):
        """
        Parameters
        ----------
        node_features: np.ndarray
            Node feature matrix with shape [num_nodes, num_node_features]
        edge_index: np.ndarray, dtype int
            Graph connectivity in COO format with shape [2, num_edges]
        edge_features: np.ndarray, optional (default None)
            Edge feature matrix with shape [num_edges, num_edge_features]
        node_pos_features: np.ndarray, optional (default None)
            Node position matrix with shape [num_nodes, num_dimensions].
        kwargs: optional
            Additional attributes and their values
        """
        # validate params
        if isinstance(node_features, np.ndarray) is False:
            raise ValueError('node_features must be np.ndarray.')

        if isinstance(edge_index, np.ndarray) is False:
            raise ValueError('edge_index must be np.ndarray.')
        elif issubclass(edge_index.dtype.type, np.integer) is False:
            raise ValueError('edge_index.dtype must contains integers.')
        elif edge_index.shape[0] != 2:
            raise ValueError('The shape of edge_index is [2, num_edges].')

        # np.max() method works only for a non-empty array, so size of the array should be non-zero
        elif (edge_index.size != 0) and (np.max(edge_index) >=
                                         len(node_features)):
            raise ValueError('edge_index contains the invalid node number.')

        if edge_features is not None:
            if isinstance(edge_features, np.ndarray) is False:
                raise ValueError('edge_features must be np.ndarray or None.')
            elif edge_index.shape[1] != edge_features.shape[0]:
                raise ValueError(
                    'The first dimension of edge_features must be the \
                          same as the second dimension of edge_index.')

        if node_pos_features is not None:
            if isinstance(node_pos_features, np.ndarray) is False:
                raise ValueError(
                    'node_pos_features must be np.ndarray or None.')
            elif node_pos_features.shape[0] != node_features.shape[0]:
                raise ValueError(
                    'The length of node_pos_features must be the same as the \
                          length of node_features.')

        self.node_features = node_features
        self.edge_index = edge_index
        self.edge_features = edge_features
        self.node_pos_features = node_pos_features
        self.kwargs = kwargs
        self.num_nodes, self.num_node_features = self.node_features.shape
        self.num_edges = edge_index.shape[1]
        if self.edge_features is not None:
            self.num_edge_features = self.edge_features.shape[1]

        for key, value in self.kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        """Returns a string containing the printable representation of the object"""
        cls = self.__class__.__name__
        node_features_str = str(list(self.node_features.shape))
        edge_index_str = str(list(self.edge_index.shape))
        if self.edge_features is not None:
            edge_features_str = str(list(self.edge_features.shape))
        else:
            edge_features_str = "None"

        out = "%s(node_features=%s, edge_index=%s, edge_features=%s" % (
            cls, node_features_str, edge_index_str, edge_features_str)
        # Adding shapes of kwargs
        for key, value in self.kwargs.items():
            if isinstance(value, np.ndarray):
                out += (', ' + key + '=' + str(list(value.shape)))
            elif isinstance(value, str):
                out += (', ' + key + '=' + value)
            elif isinstance(value, int) or isinstance(value, float):
                out += (', ' + key + '=' + str(value))
        out += ')'
        return out

    def to_pyg_graph(self):
        """Convert to PyTorch Geometric graph data instance

        Returns
        -------
        torch_geometric.data.Data
            Graph data for PyTorch Geometric

        Note
        ----
        This method requires PyTorch Geometric to be installed.
        """
        try:
            import torch
            from torch_geometric.data import Data
        except ModuleNotFoundError:
            raise ImportError(
                "This function requires PyTorch Geometric to be installed.")

        edge_features = self.edge_features
        if edge_features is not None:
            edge_features = torch.from_numpy(self.edge_features).float()
        node_pos_features = self.node_pos_features
        if node_pos_features is not None:
            node_pos_features = torch.from_numpy(self.node_pos_features).float()
        kwargs = {}
        for key, value in self.kwargs.items():
            kwargs[key] = torch.from_numpy(value).float()
        return Data(x=torch.from_numpy(self.node_features).float(),
                    edge_index=torch.from_numpy(self.edge_index).long(),
                    edge_attr=edge_features,
                    pos=node_pos_features,
                    **kwargs)

    def to_dgl_graph(self, self_loop: bool = False):
        """Convert to DGL graph data instance

        Returns
        -------
        dgl.DGLGraph
            Graph data for DGL
        self_loop: bool
            Whether to add self loops for the nodes, i.e. edges from nodes
            to themselves. Default to False.

        Note
        ----
        This method requires DGL to be installed.
        """
        try:
            import dgl
            import torch
        except ModuleNotFoundError:
            raise ImportError("This function requires DGL to be installed.")

        src = self.edge_index[0]
        dst = self.edge_index[1]

        g = dgl.graph(
            (torch.from_numpy(src).long(), torch.from_numpy(dst).long()),
            num_nodes=self.num_nodes)
        g.ndata['x'] = torch.from_numpy(self.node_features).float()

        if self.node_pos_features is not None:
            g.ndata['pos'] = torch.from_numpy(self.node_pos_features).float()

        if self.edge_features is not None:
            g.edata['edge_attr'] = torch.from_numpy(self.edge_features).float()

        if self_loop:
            # This assumes that the edge features for self loops are full-zero tensors
            # In the future we may want to support featurization for self loops
            g.add_edges(np.arange(self.num_nodes), np.arange(self.num_nodes))

        return g

    def numpy_to_torch(self, device: str = 'cpu'):
        """Convert numpy arrays to torch tensors. This may be useful when you are using PyTorch Geometric with GraphData objects.

        Parameters
        ----------
        device : str
            Device to store the tensors. Default to 'cpu'.

        Example
        -------
        >>> num_nodes, num_node_features = 5, 32
        >>> num_edges, num_edge_features = 6, 32
        >>> node_features = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data = GraphData(node_features, edge_index, edge_features)
        >>> graph_data = graph_data.numpy_to_torch()
        >>> print(type(graph_data.node_features))
        <class 'torch.Tensor'>
        """
        import copy

        import torch
        graph_copy = copy.deepcopy(self)

        graph_copy.node_features = torch.from_numpy(  # type: ignore
            self.node_features).float().to(device)
        graph_copy.edge_index = torch.from_numpy(  # type: ignore
            self.edge_index).long().to(device)
        if self.edge_features is not None:
            graph_copy.edge_features = torch.from_numpy(  # type: ignore
                self.edge_features).float().to(device)
        else:
            graph_copy.edge_features = None
        if self.node_pos_features is not None:
            graph_copy.node_pos_features = torch.from_numpy(  # type: ignore
                self.node_pos_features).float().to(device)
        else:
            graph_copy.node_pos_features = None

        graph_copy.kwargs = {}
        for key, value in self.kwargs.items():
            if isinstance(value, np.ndarray):
                value = torch.from_numpy(value).to(device)
                graph_copy.kwargs[key] = value
                setattr(graph_copy, key, value)

        return graph_copy


class BatchGraphData(GraphData):
    """Batch GraphData class

    Attributes
    ----------
    node_features: np.ndarray
        Concatenated node feature matrix with shape [num_nodes, num_node_features].
        `num_nodes` is total number of nodes in the batch graph.
    edge_index: np.ndarray, dtype int
        Concatenated graph connectivity in COO format with shape [2, num_edges].
        `num_edges` is total number of edges in the batch graph.
    edge_features: np.ndarray, optional (default None)
        Concatenated edge feature matrix with shape [num_edges, num_edge_features].
        `num_edges` is total number of edges in the batch graph.
    node_pos_features: np.ndarray, optional (default None)
        Concatenated node position matrix with shape [num_nodes, num_dimensions].
        `num_nodes` is total number of edges in the batch graph.
    num_nodes: int
        The number of nodes in the batch graph.
    num_node_features: int
        The number of features per node in the graph.
    num_edges: int
        The number of edges in the batch graph.
    num_edges_features: int, optional (default None)
        The number of features per edge in the graph.
    graph_index: np.ndarray, dtype int
        This vector indicates which graph the node belongs with shape [num_nodes,].

    Examples
    --------
    >>> import numpy as np
    >>> from deepchem.feat.graph_data import GraphData
    >>> node_features_list = np.random.rand(2, 5, 10)
    >>> edge_index_list = np.array([
    ...    [[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]],
    ...    [[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]],
    ... ], dtype=int)
    >>> graph_list = [GraphData(node_features, edge_index) for node_features, edge_index
    ...           in zip(node_features_list, edge_index_list)]
    >>> batch_graph = BatchGraphData(graph_list=graph_list)
    """

    def __init__(self, graph_list: Sequence[GraphData]):
        """
        Parameters
        ----------
        graph_list: Sequence[GraphData]
            List of GraphData
        """
        # stack features
        batch_node_features = np.vstack(
            [graph.node_features for graph in graph_list])

        # before stacking edge_features or node_pos_features,
        # we should check whether these are None or not
        if graph_list[0].edge_features is not None:
            batch_edge_features: Optional[np.ndarray] = np.vstack(
                [graph.edge_features for graph in graph_list])  # type: ignore
        else:
            batch_edge_features = None

        if graph_list[0].node_pos_features is not None:
            batch_node_pos_features: Optional[np.ndarray] = np.vstack([
                graph.node_pos_features for graph in graph_list  # type: ignore
            ])
        else:
            batch_node_pos_features = None

        # create new edge index
        # number of nodes in each graph
        num_nodes_list = [graph.num_nodes for graph in graph_list]
        # cumulative number of nodes for each graph. This is necessary because the values in edge_index are node indices of all of the graphs in graph_list and so we need to offset the indices by the number of nodes in the previous graphs.
        cum_num_nodes_list = np.cumsum([0] + num_nodes_list)[:-1]
        # columns are the edge index, values are the node index
        batch_edge_index = np.hstack([
            graph.edge_index + cum_num_nodes
            for cum_num_nodes, graph in zip(cum_num_nodes_list, graph_list)
        ])

        # graph_index indicates which nodes belong to which graph
        graph_index = []
        for i, num_nodes in enumerate(num_nodes_list):
            graph_index.extend([i] * num_nodes)
        self.graph_index = np.array(graph_index)

        super().__init__(
            node_features=batch_node_features,
            edge_index=batch_edge_index,
            edge_features=batch_edge_features,
            node_pos_features=batch_node_pos_features,
        )

    def numpy_to_torch(self, device: str = "cpu"):
        """
        Convert numpy arrays to torch tensors for BatchGraphData. BatchGraphData is very similar to GraphData, but it combines all graphs into a single graph object and it has an additional attribute `graph_index` which indicates which nodes belong to which graph.

        Parameters
        ----------
        device : str
            Device to store the tensors. Default to 'cpu'.

        Example
        -------
        >>> num_nodes, num_node_features = 5, 32
        >>> num_edges, num_edge_features = 6, 32
        >>> node_features = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data = GraphData(node_features, edge_index, edge_features)
        >>> node_features2 = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features2 = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index2 = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data2 = GraphData(node_features2, edge_index2, edge_features2)
        >>> batch_graph_data = BatchGraphData([graph_data, graph_data2])
        >>> batch_graph_data = batch_graph_data.numpy_to_torch()
        >>> print(type(batch_graph_data.node_features))
        <class 'torch.Tensor'>
        """
        import torch
        graph_copy = super().numpy_to_torch(device)

        graph_index = torch.from_numpy(
            self.graph_index).long().to(  # type: ignore
                device)
        graph_copy.graph_index = graph_index

        return graph_copy



"""
Copied from https://github.com/tencent-ailab/grover/blob/0421d97a5e1bd1b59d1923e3afd556afbe4ff782/grover/model/layers.py
"""
from typing import List, Dict
try:
    import torch
    import torch.nn as nn
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')
import numpy as np
from scipy import stats
from deepchem.models.torch_models.readout import GroverReadout
from deepchem.models.torch_models.layers import SublayerConnection, PositionwiseFeedForward


class GroverEmbedding(nn.Module):
    """GroverEmbedding layer.

    This layer is a simple wrapper over GroverTransEncoder layer for retrieving the embeddings from the GroverTransEncoder layer.

    Parameters
    ----------
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    node_fdim: int
        the dimension of additional feature for node/atom.
    depth: int
        Dynamic message passing depth for use in MPNEncoder
    undirected: bool
        The message passing is undirected or not
    num_mt_block: int
        the number of message passing blocks.
    num_head: int
        the number of attention heads.
    """

    def __init__(self,
                 node_fdim,
                 edge_fdim,
                 hidden_size=128,
                 depth=1,
                 undirected=False,
                 dropout=0.2,
                 activation='relu',
                 num_mt_block=1,
                 num_heads=4,
                 bias=False,
                 res_connection=False):
        super(GroverEmbedding, self).__init__()
        self.encoders = GroverTransEncoder(hidden_size=hidden_size,
                                           edge_fdim=edge_fdim,
                                           node_fdim=node_fdim,
                                           depth=depth,
                                           undirected=undirected,
                                           dropout=dropout,
                                           activation=activation,
                                           num_mt_block=num_mt_block,
                                           num_heads=num_heads,
                                           bias=bias,
                                           res_connection=res_connection)

    def forward(self, graph_batch: List[torch.Tensor]):
        """Forward function

        Parameters
        ----------
        graph_batch: List[torch.Tensor]
            A list containing f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a

        Returns
        -------
        embedding: Dict[str, torch.Tensor]
            When embedding type is `atom`, the dictionary contains atom and bond embeddings aggregated by atom, if `bond`, the dictionary contains atom and bond embedding aggregated by bond and if `both`, it contains aotm and bond embeddings aggregated to both atom and bond.
        """
        output = self.encoders(graph_batch)
        return {
            "atom_from_atom": output[0][0],
            "bond_from_atom": output[0][1],
            "atom_from_bond": output[1][0],
            "bond_from_bond": output[1][1]
        }


class GroverBondVocabPredictor(nn.Module):
    """Layer for learning contextual information for bonds.

    The layer is used in Grover architecture to learn contextual information of a bond by predicting
    the context of a bond from the bond embedding in a multi-class classification setting.
    The contextual information of a bond are encoded as strings (ex: '(DOUBLE-STEREONONE-NONE)_C-(SINGLE-STEREONONE-NONE)2').

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverBondVocabPredictor
    >>> num_bonds = 20
    >>> in_features, vocab_size = 16, 10
    >>> layer = GroverBondVocabPredictor(vocab_size, in_features)
    >>> embedding = torch.randn(num_bonds * 2 + 1,
    ...  in_features)  # * 2 + 1 for reverse bond and padding
    >>> result = layer(embedding)
    >>> result.shape
    torch.Size([21, 10])

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, vocab_size: int, in_features: int = 128):
        """Initializes GroverBondVocabPredictor

        Parameters
        ----------
        vocab_size: int
            Size of vocabulary, used for number of classes in prediction.
        in_features: int, default: 128
            Input feature size of bond embeddings.
        """
        super(GroverBondVocabPredictor, self).__init__()
        self.linear = nn.Linear(in_features, vocab_size)
        self.linear_rev = nn.Linear(in_features, vocab_size)
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, embeddings):
        """
        Parameters
        ----------
        embeddings: torch.Tensor
            bond embeddings of shape (num_bond, in_features)

        Returns
        -------
        logits: torch.Tensor
            the prediction for each bond, (num_bond, vocab_size)
        """
        nm_bonds = embeddings.shape[
            0]  # must be an odd number because every bond (undirected edge) is represented as two directed edges + 1 element for padding.
        # The bond and rev bond have odd and even ids respectively.
        ids1 = [0] + list(range(1, nm_bonds, 2))
        ids2 = list(range(0, nm_bonds, 2))
        logits = self.linear(embeddings[ids1]) + self.linear_rev(
            embeddings[ids2])
        return self.logsoftmax(logits)


class GroverAtomVocabPredictor(nn.Module):
    """Grover Atom Vocabulary Prediction Module.

    The GroverAtomVocabPredictor module is used for predicting atom-vocabulary
    for the self-supervision task in Grover architecture. In the self-supervision tasks,
    one task is to learn contextual-information of nodes (atoms).
    Contextual information are encoded as strings, like `C_N-DOUBLE1_O-SINGLE1`.
    The module accepts an atom encoding and learns to predict the contextual information
    of the atom as a multi-class classification problem.

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverAtomVocabPredictor
    >>> num_atoms, in_features, vocab_size = 30, 16, 10
    >>> layer = GroverAtomVocabPredictor(vocab_size, in_features)
    >>> embedding = torch.randn(num_atoms, in_features)
    >>> result = layer(embedding)
    >>> result.shape
    torch.Size([30, 10])

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, vocab_size: int, in_features: int = 128):
        """Initializing Grover Atom Vocabulary Predictor

        Parameters
        ----------
        vocab_size: int
            size of vocabulary (vocabulary here is the total number of different possible contexts)
        in_features: int
            feature size of atom embeddings.
        """
        super(GroverAtomVocabPredictor, self).__init__()
        self.linear = nn.Linear(in_features, vocab_size)
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, embeddings):
        """
        Parameters
        ----------
        embeddings: torch.Tensor
            the atom embeddings of shape (vocab_size, in_features)

        Returns
        -------
        logits: torch.Tensor
            the prediction for each atom of shape (num_bond, vocab_size)
        """
        return self.logsoftmax(self.linear(embeddings))


class GroverFunctionalGroupPredictor(nn.Module):
    """The functional group prediction task for self-supervised learning.

    Molecules have functional groups in them. This module is used for predicting
    the functional group and the problem is formulated as an multi-label classification problem.

    Parameters
    ----------
    fg_size: int,
        size of functional group
    in_features: int,
        hidden_layer size, default 128

    Example
    -------
    >>> from deepchem.models.torch_models.grover_layers import GroverFunctionalGroupPredictor
    >>> in_features, fg_size = 8, 20
    >>> num_atoms, num_bonds = 10, 20
    >>> predictor = GroverFunctionalGroupPredictor(fg_size=20, in_features=8)
    >>> atom_scope, bond_scope = [(0, 3), (3, 3), (6, 4)], [(0, 5), (5, 4), (9, 11)]
    >>> embeddings = {}
    >>> embeddings['bond_from_atom'] = torch.randn(num_bonds, in_features)
    >>> embeddings['bond_from_bond'] = torch.randn(num_bonds, in_features)
    >>> embeddings['atom_from_atom'] = torch.randn(num_atoms, in_features)
    >>> embeddings['atom_from_bond'] = torch.randn(num_atoms, in_features)
    >>> result = predictor(embeddings, atom_scope, bond_scope)

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.

    """

    def __init__(self, fg_size: int, in_features=128):
        super(GroverFunctionalGroupPredictor, self).__init__()

        self.readout = GroverReadout(rtype="mean", in_features=in_features)
        self.linear_atom_from_atom = nn.Linear(in_features, fg_size)
        self.linear_atom_from_bond = nn.Linear(in_features, fg_size)
        self.linear_bond_from_atom = nn.Linear(in_features, fg_size)
        self.linear_bond_from_bond = nn.Linear(in_features, fg_size)

    def forward(self, embeddings: Dict, atom_scope: List, bond_scope: List):
        """
        The forward function for the GroverFunctionalGroupPredictor (semantic motif prediction) layer.
        It takes atom/bond embeddings produced from GroverEmbedding module and
        their corresponsing scopes and produces prediction logits for different branches.
        The scopes are used to differentiate atoms/bonds belonging to a molecule in a
        batched molecular graph.

        Parameters
        ----------
        embedding: Dict
            The input embeddings organized as an dictionary. The input embeddings are output of GroverEmbedding layer.
        atom_scope: List
            The scope for atoms.
        bond_scope: List
            The scope for bonds

        Returns
        -------
        preds: Dict
            A dictionary containing the predicted logits.
        """
        preds_atom_from_atom, preds_atom_from_bond, preds_bond_from_atom, preds_bond_from_bond = None, None, None, None

        if embeddings["bond_from_atom"] is not None:
            preds_bond_from_atom = self.linear_bond_from_atom(
                self.readout(embeddings["bond_from_atom"], bond_scope))
        if embeddings["bond_from_bond"] is not None:
            preds_bond_from_bond = self.linear_bond_from_bond(
                self.readout(embeddings["bond_from_bond"], bond_scope))

        if embeddings["atom_from_atom"] is not None:
            preds_atom_from_atom = self.linear_atom_from_atom(
                self.readout(embeddings["atom_from_atom"], atom_scope))
        if embeddings["atom_from_bond"] is not None:
            preds_atom_from_bond = self.linear_atom_from_bond(
                self.readout(embeddings["atom_from_bond"], atom_scope))

        return {
            "atom_from_atom": preds_atom_from_atom,
            "atom_from_bond": preds_atom_from_bond,
            "bond_from_atom": preds_bond_from_atom,
            "bond_from_bond": preds_bond_from_bond
        }


def _index_select_nd(source: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
    """
    Selects the message features from source corresponding to the atom or bond indices in index.

    Parameters
    ----------
    source: torch.Tensor
        A tensor of shape (num_bonds, hidden_size) containing message features.
    index: torch.Tensor
        A tensor of shape (num_atoms/num_bonds, max_num_bonds) containing the atom or bond indices to select from source.

    Returns
    ----------
    message_features: torch.Tensor
        A tensor of shape (num_atoms/num_bonds, max_num_bonds, hidden_size) containing the message features corresponding to the atoms/bonds specified in index.
    """
    index_size = index.size()  # (num_atoms/num_bonds, max_num_bonds)
    suffix_dim = source.size()[1:]  # (hidden_size,)
    final_size = index_size + suffix_dim  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    target = source.index_select(dim=0, index=index.view(
        -1))  # (num_atoms/num_bonds * max_num_bonds, hidden_size)
    target = target.view(
        final_size)  # (num_atoms/num_bonds, max_num_bonds, hidden_size)

    return target


def _select_neighbor_and_aggregate(feature, index):
    """The basic operation in message passing.

    Caution: the index_selec_ND would cause the reproducibility issue when performing the training on CUDA.

    The operation is like map and reduce. `index_select_nd` maps message features to bonds/atoms and
    this method aggregates the results by using `sum` as pooling operation. We can also add a configuration
    here to use `mean` as pooling operation but it is left to future implementation.

    References
    ----------
    See: https://pytorch.org/docs/stable/notes/randomness.html

    Parameters
    ----------
    feature: np.array
        The candidate feature for aggregate. (n_nodes, hidden)
    index: np.array
        The selected index (neighbor indexes).

    Returns
    -------
    None
    """
    neighbor = _index_select_nd(feature, index)
    return neighbor.sum(dim=1)


class GroverMPNEncoder(nn.Module):
    """Performs Message Passing to generate encodings for the molecule.

    Parameters
    ----------
    atom_messages: bool
        True if encoding atom-messages else False.
    init_message_dim: int
        Dimension of embedding message.
    attach_feats: bool
        Set to `True` if additional features are passed along with node/edge embeddings.
    attached_feat_fdim: int
        Dimension of additional features when `attach_feats` is `True`
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    depth: int
        number of hops in a message passing iteration
    dynamic_depth: str, default: none
        If set to `uniform` for randomly sampling dynamic depth from an uniform distribution else if set to `truncnorm`, dynamic depth is sampled from a truncated normal distribution.
    input_layer: str
        If set to `fc`, adds an initial feed-forward layer. If set to `none`, does not add an initial feed forward layer.
    """

    # FIXME This layer is similar to DMPNNEncoderLayer and they
    # must be unified.
    def __init__(self,
                 atom_messages: bool,
                 init_message_dim: int,
                 hidden_size: int,
                 depth: int,
                 undirected: bool,
                 attach_feats: bool,
                 attached_feat_fdim: int = 0,
                 bias: bool = True,
                 dropout: float = 0.2,
                 activation: str = 'relu',
                 input_layer: str = 'fc',
                 dynamic_depth: str = 'none'):
        super(GroverMPNEncoder, self).__init__()
        if input_layer == 'none':
            assert init_message_dim == hidden_size
        assert dynamic_depth in [
            'uniform', 'truncnorm', 'none'
        ], 'dynamic depth should be one of uniform, truncnorm, none'
        self.init_message_dim = init_message_dim
        self.depth = depth
        self.input_layer = input_layer
        self.layers_per_message = 1
        self.undirected = undirected
        self.atom_messages = atom_messages
        self.attached_feat = attach_feats

        assert dynamic_depth in [
            'none', 'truncnorm', 'uniform'
        ], 'If dynamic depth, it should be truncnorm or uniform'
        self.dynamic_depth = dynamic_depth

        self.dropout_layer = nn.Dropout(p=dropout)
        if activation == 'relu':
            self.act_func = nn.ReLU()
        else:
            raise ValueError('Only ReLU activation function is supported')

        if self.input_layer == "fc":
            self.W_i = nn.Linear(self.init_message_dim, hidden_size, bias=bias)
            w_h_input_size = hidden_size
        elif input_layer == 'none':
            w_h_input_size = self.init_message_dim

        if self.attached_feat:
            w_h_input_size += attached_feat_fdim

        # Shared weight matrix across depths (default)
        self.W_h = nn.Linear(w_h_input_size, hidden_size, bias=bias)

    def forward(self,
                init_messages,
                init_attached_features,
                a2nei,
                a2attached,
                b2a=None,
                b2revb=None,
                adjs=None) -> torch.FloatTensor:
        if self.input_layer == 'fc':
            message = self.act_func(self.W_i(init_messages))
        elif self.input_layer == 'none':
            message = init_messages

        attached_feats = init_attached_features

        if self.training and self.dynamic_depth != 'none':
            if self.dynamic_depth == 'uniform':
                ndepth = abs(
                    int(np.random.uniform(self.depth - 3, self.depth + 3)))
            elif self.dynamic_depth == 'truncnorm':
                mu, sigma = self.depth, 1
                lower, upper = mu - 3 * sigma, mu + 3 * sigma
                X = stats.truncnorm((lower - mu) / sigma, (upper - mu) / sigma,
                                    loc=mu,
                                    scale=sigma)
                ndepth = int(X.rvs(1))
        else:
            ndepth = self.depth

        for _ in range(ndepth - 1):
            if self.undirected:
                message = (message + message[b2revb]) / 2

            nei_message = _select_neighbor_and_aggregate(message, a2nei)
            a_message = nei_message
            if self.attached_feat:
                attached_nei_feats = _select_neighbor_and_aggregate(
                    attached_feats, a2attached)
                a_message = torch.cat((nei_message, attached_nei_feats), dim=1)

            if not self.atom_messages:
                rev_message = message[b2revb]
                if self.attached_feat:
                    atom_rev_message = attached_feats[b2a[b2revb]]
                    rev_message = torch.cat((rev_message, atom_rev_message),
                                            dim=1)
                # Except reverse bond its-self(w) ! \sum_{k\in N(u) \ w}
                message = a_message[b2a] - rev_message  # num_bonds x hidden
            else:
                message = a_message

            # FIXME When input_layer is none, for the first iteration of message passing, we should ideally
            # be using different weight matrix since message will be of shape num_bonds x f_bonds_dim
            # in the first iteration and in the subsequent iterations, it will be num_bonds x hidden_size
            message = self.W_h(message)

            message = self.dropout_layer(self.act_func(message))

        return message


class GroverAttentionHead(nn.Module):
    """Generates attention head using GroverMPNEncoder for generating query, key and value

    Parameters
    ----------
    hidden_size: int
        Dimension of hidden layer
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    depth: int
        number of hops in a message passing iteration
    atom_messages: bool
        True if encoding atom-messages else False.
    """

    def __init__(self,
                 hidden_size: int = 128,
                 bias: bool = True,
                 depth: int = 1,
                 dropout: float = 0.0,
                 undirected: bool = False,
                 atom_messages: bool = False):
        super(GroverAttentionHead, self).__init__()
        self.atom_messages = atom_messages

        # FIXME We assume that we are using a hidden layer to transform the initial atom message
        # and bond messages to hidden dimension size.
        self.mpn_q = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

        self.mpn_k = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

        self.mpn_v = GroverMPNEncoder(atom_messages=atom_messages,
                                      init_message_dim=hidden_size,
                                      attached_feat_fdim=hidden_size,
                                      hidden_size=hidden_size,
                                      bias=bias,
                                      depth=depth,
                                      dropout=dropout,
                                      undirected=undirected,
                                      attach_feats=False,
                                      input_layer='none',
                                      dynamic_depth='truncnorm')

    def forward(self, f_atoms, f_bonds, a2b, a2a, b2a, b2revb):
        if self.atom_messages:
            init_messages = f_atoms
            init_attached_features = f_bonds
            a2nei = a2a
            a2attached = a2b
            b2a = b2a
            b2revb = b2revb
        else:
            # self.atom_messages is False
            init_messages = f_bonds
            init_attached_features = f_atoms
            a2nei = a2b
            a2attached = a2a
            b2a = b2a
            b2revb = b2revb

        q = self.mpn_q(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        k = self.mpn_k(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        v = self.mpn_v(init_messages=init_messages,
                       init_attached_features=init_attached_features,
                       a2nei=a2nei,
                       a2attached=a2attached,
                       b2a=b2a,
                       b2revb=b2revb)
        return q, k, v


class GroverMTBlock(nn.Module):
    """Message passing combined with transformer architecture

    The layer combines message passing performed using GroverMPNEncoder and uses it
    to generate query, key and value for multi-headed Attention block.

    Parameters
    ----------
    atom_messages: bool
        True if encoding atom-messages else False.
    input_dim: int
        Dimension of input features
    num_heads: int
        Number of attention heads
    depth: int
        Number of hops in a message passing iteration
    undirected: bool
        If set to `True`, the graph is considered as an undirected graph.
    """

    def __init__(self,
                 atom_messages: bool,
                 input_dim: int,
                 num_heads: int,
                 depth: int,
                 undirected: bool = False,
                 hidden_size: int = 128,
                 dropout: float = 0.0,
                 bias: bool = True,
                 res_connection: bool = True,
                 activation: str = 'relu'):
        super(GroverMTBlock, self).__init__()
        self.hidden_size = hidden_size
        self.atom_messages = atom_messages
        self.res_connection = res_connection
        self.num_heads = num_heads
        if activation == 'relu':
            self.act_func = nn.ReLU()
        else:
            raise ValueError('Only relu activation is supported')
        self.dropout_layer = nn.Dropout(p=dropout)

        # Note: Elementwise affine has to be consistent with the pre-training phase
        self.layernorm = nn.LayerNorm(hidden_size, elementwise_affine=True)
        self.attn = torch.nn.MultiheadAttention(embed_dim=hidden_size,
                                                num_heads=num_heads,
                                                dropout=dropout,
                                                bias=bias,
                                                batch_first=True)

        self.W_i = nn.Linear(input_dim, hidden_size, bias)
        self.W_o = nn.Linear(hidden_size * num_heads, hidden_size, bias)
        self.sublayer = SublayerConnection(hidden_size, dropout)
        self.attention_heads = nn.ModuleList()
        for i in range(num_heads):
            self.attention_heads.append(
                GroverAttentionHead(hidden_size=hidden_size,
                                    bias=bias,
                                    dropout=dropout,
                                    depth=depth,
                                    atom_messages=atom_messages,
                                    undirected=undirected))

    def forward(self, batch):
        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a = batch
        if self.atom_messages:
            if f_atoms.shape[1] != self.hidden_size:
                f_atoms = self.W_i(f_atoms)
                f_atoms = self.dropout_layer(
                    self.layernorm(self.act_func(f_atoms)))

        else:
            if f_bonds.shape[1] != self.hidden_size:
                f_bonds = self.W_i(f_bonds)
                f_bonds = self.dropout_layer(
                    self.layernorm(self.act_func(f_bonds)))

        queries, keys, values = [], [], []
        for head in self.attention_heads:
            q, k, v = head(f_atoms,
                           f_bonds,
                           a2b=a2b,
                           b2a=b2a,
                           b2revb=b2revb,
                           a2a=a2a)
            queries.append(q.unsqueeze(1))
            keys.append(k.unsqueeze(1))
            values.append(v.unsqueeze(1))
        queries = torch.cat(queries, dim=1)
        keys = torch.cat(keys, dim=1)
        values = torch.cat(values, dim=1)

        # multi-headed attention
        x_out, _ = self.attn(queries, keys, values)
        x_out = x_out.reshape(x_out.shape[0], -1)
        x_out = self.W_o(x_out)

        # support no residual connection in MTBlock.
        if self.res_connection:
            if self.atom_messages:
                f_atoms = self.sublayer(f_atoms, x_out)
            else:
                f_bonds = self.sublayer(f_bonds, x_out)
        batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
        return batch


class GroverTransEncoder(nn.Module):
    """GroverTransEncoder for encoding a molecular graph

    The GroverTransEncoder layer is used for encoding a molecular graph.
    The layer can return four possible output depending on the `atom_emb_output`
    choice. If it `none`, it returns output from multihead attention block directly.
    If it is `atom`, then it aggregates node messages from node hidden states, it also aggregates
    node messages from incoming bond's hidden states to the nodes and returns two tensors.
    It it is `bond`, it aggregates bond embeddings from bond hidden states, it also aggregates another
    set of bond embeddings from bond's source nodes hidden states and returns two tensors.
    If it is `both`, it returns four tensor (`bond` option + `atom` option).

    Parameters
    ----------
    hidden_size: int
        the hidden size of the model.
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    node_fdim: int
        the dimension of additional feature for node/atom.
    depth: int
        Dynamic message passing depth for use in MPNEncoder
    undirected: bool
        The message passing is undirected or not
    dropout: float
        the dropout ratio
    activation: str
        the activation function
    num_mt_block: int
        the number of mt block.
    num_head: int
        the number of attention AttentionHead.
    bias: bool
        enable bias term in all linear layers.
    res_connection: bool
        enables the skip-connection in MTBlock.
    """

    def __init__(self,
                 node_fdim: int,
                 edge_fdim: int,
                 depth: int = 3,
                 undirected: bool = False,
                 num_mt_block: int = 2,
                 num_heads: int = 2,
                 hidden_size: int = 64,
                 dropout: float = 0.2,
                 res_connection: bool = True,
                 bias: bool = True,
                 activation: str = 'relu'):
        super(GroverTransEncoder, self).__init__()

        self.hidden_size = hidden_size
        self.edge_fdim = edge_fdim
        self.node_fdim = node_fdim

        self.edge_blocks = nn.ModuleList()
        self.node_blocks = nn.ModuleList()

        for i in range(num_mt_block):
            if i == 0:
                node_input_fdim, edge_input_fdim = node_fdim, edge_fdim
            else:
                node_input_fdim, edge_input_fdim = hidden_size, hidden_size

            self.edge_blocks.append(
                GroverMTBlock(num_heads=num_heads,
                              input_dim=edge_input_fdim,
                              hidden_size=hidden_size,
                              activation=activation,
                              dropout=dropout,
                              bias=bias,
                              atom_messages=False,
                              res_connection=res_connection,
                              depth=depth,
                              undirected=undirected))
            self.node_blocks.append(
                GroverMTBlock(num_heads=num_heads,
                              input_dim=node_input_fdim,
                              hidden_size=hidden_size,
                              activation=activation,
                              dropout=dropout,
                              bias=bias,
                              atom_messages=True,
                              res_connection=res_connection,
                              depth=depth,
                              undirected=undirected))

        self.ffn_atom_from_atom = PositionwiseFeedForward(
            d_input=self.hidden_size + node_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_atom_from_bond = PositionwiseFeedForward(
            d_input=self.hidden_size + node_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_bond_from_atom = PositionwiseFeedForward(
            d_input=self.hidden_size + edge_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.ffn_bond_from_bond = PositionwiseFeedForward(
            d_input=self.hidden_size + edge_fdim,
            d_hidden=self.hidden_size * 4,
            d_output=self.hidden_size,
            n_layers=2,
            activation=activation,
            dropout_p=dropout)

        self.atom_from_atom_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.atom_from_bond_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.bond_from_atom_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)
        self.bond_from_bond_sublayer = SublayerConnection(size=self.hidden_size,
                                                          dropout_p=dropout)

        if activation == 'relu':
            self.act_func_node = nn.ReLU()
            self.act_func_edge = nn.ReLU()
        else:
            raise ValueError('Only relu activation is supported')

        self.dropout_layer = nn.Dropout(p=dropout)

    def _pointwise_feed_forward_to_atom_embedding(self, emb_output, atom_feat,
                                                  index, ffn_layer):
        aggr_output = _select_neighbor_and_aggregate(emb_output, index)
        aggr_outputx = torch.cat([atom_feat, aggr_output], dim=1)
        return ffn_layer(aggr_outputx), aggr_outputx

    def _pointwise_feed_forward_to_bond_embedding(self, emb_output, bond_feat,
                                                  a2nei, b2revb, ffn_layer):
        aggr_output = _select_neighbor_and_aggregate(emb_output, a2nei)
        aggr_output = self._remove_rev_bond_message(emb_output, aggr_output,
                                                    b2revb)
        aggr_outputx = torch.cat([bond_feat, aggr_output], dim=1)
        return ffn_layer(aggr_outputx), aggr_outputx

    @staticmethod
    def _remove_rev_bond_message(original_message, aggr_message, b2revb):
        rev_message = original_message[b2revb]
        return aggr_message - rev_message

    def _atom_bond_transform(
            self,
            to_atom=True,  # False: to bond
            atomwise_input=None,
            bondwise_input=None,
            original_f_atoms=None,
            original_f_bonds=None,
            a2a=None,
            a2b=None,
            b2a=None,
            b2revb=None):
        """Transfer the output of atom/bond multi-head attention to the final atom/bond output.
        """

        if to_atom:
            # atom input to atom output
            atomwise_input, _ = self._pointwise_feed_forward_to_atom_embedding(
                atomwise_input, original_f_atoms, a2a, self.ffn_atom_from_atom)
            atom_in_atom_out = self.atom_from_atom_sublayer(
                None, atomwise_input)
            # bond to atom
            bondwise_input, _ = self._pointwise_feed_forward_to_atom_embedding(
                bondwise_input, original_f_atoms, a2b, self.ffn_atom_from_bond)
            bond_in_atom_out = self.atom_from_bond_sublayer(
                None, bondwise_input)
            return atom_in_atom_out, bond_in_atom_out
        else:  # to bond embeddings
            # atom input to bond output
            atom_list_for_bond = torch.cat([b2a.unsqueeze(dim=1), a2a[b2a]],
                                           dim=1)
            atomwise_input, _ = self._pointwise_feed_forward_to_bond_embedding(
                atomwise_input, original_f_bonds, atom_list_for_bond,
                b2a[b2revb], self.ffn_bond_from_atom)
            atom_in_bond_out = self.bond_from_atom_sublayer(
                None, atomwise_input)
            # bond input to bond output
            bond_list_for_bond = a2b[b2a]
            bondwise_input, _ = self._pointwise_feed_forward_to_bond_embedding(
                bondwise_input, original_f_bonds, bond_list_for_bond, b2revb,
                self.ffn_bond_from_bond)
            bond_in_bond_out = self.bond_from_bond_sublayer(
                None, bondwise_input)
            return atom_in_bond_out, bond_in_bond_out

    def forward(self, batch):
        f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a = batch

        node_batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
        edge_batch = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a

        original_f_atoms, original_f_bonds = f_atoms, f_bonds

        for nb in self.node_blocks:  # atom messages. Multi-headed attention
            node_batch = nb(node_batch)
        for eb in self.edge_blocks:  # bond messages. Multi-headed attention
            edge_batch = eb(edge_batch)

        atom_output, _, _, _, _, _, _, _ = node_batch  # atom hidden states
        _, bond_output, _, _, _, _, _, _ = edge_batch  # bond hidden states

        atom_embeddings = self._atom_bond_transform(
            to_atom=True,  # False: to bond
            atomwise_input=atom_output,
            bondwise_input=bond_output,
            original_f_atoms=original_f_atoms,
            original_f_bonds=original_f_bonds,
            a2a=a2a,
            a2b=a2b,
            b2a=b2a,
            b2revb=b2revb)

        bond_embeddings = self._atom_bond_transform(
            to_atom=False,  # False: to bond
            atomwise_input=atom_output,
            bondwise_input=bond_output,
            original_f_atoms=original_f_atoms,
            original_f_bonds=original_f_bonds,
            a2a=a2a,
            a2b=a2b,
            b2a=b2a,
            b2revb=b2revb)

        return ((atom_embeddings[0], bond_embeddings[0]), (atom_embeddings[1],
                                                           bond_embeddings[1]))



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import GRU, Linear, ReLU, Sequential

from deepchem.feat.graph_data import BatchGraphData
from deepchem.models.losses import (
    GlobalMutualInformationLoss,
    LocalMutualInformationLoss,
)
from deepchem.models.torch_models.layers import MultilayerPerceptron
from deepchem.models.torch_models.modular import ModularTorchModel

try:
    from torch_geometric.nn import GINConv, NNConv, global_add_pool
    from torch_geometric.nn.aggr import Set2Set
except ImportError:
    pass


class GINEncoder(torch.nn.Module):
    """
    Graph Information Network (GIN) encoder. This is a graph convolutional network that produces encoded representations for molecular graph inputs.

    Parameters
    ----------
    num_features: int
        The number of node features
    embedding_dim: int
        The dimension of the output embedding
    num_gc_layers: int, optional (default 5)
        The number of graph convolutional layers to use

    Example
    -------
    >>> import numpy as np
    >>> from deepchem.models.torch_models.infograph import GINEncoder
    >>> from deepchem.feat.graph_data import GraphData
    >>> encoder = GINEncoder(num_features=25, embedding_dim=32)
    >>> node_features = np.random.randn(10, 25)
    >>> edge_index = np.array([[0, 1, 2], [1, 2, 3]])
    >>> edge_features = np.random.randn(3, 10)
    >>> graph_index = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    >>> data = GraphData(node_features=node_features, edge_index=edge_index, edge_features=edge_features, graph_index=graph_index)
    >>> data.node_features = torch.from_numpy(data.node_features).float()
    >>> data.edge_features = torch.from_numpy(data.edge_features).float()
    >>> data.edge_index = torch.from_numpy(data.edge_index).long()
    >>> data.graph_index = torch.from_numpy(data.graph_index).long()
    >>> embedding, intermediate_embeddings = encoder(data)
    >>> print(embedding.shape)
    torch.Size([1, 30])

    References
    ----------
    .. [1] Xu, K., Hu, W., Leskovec, J. & Jegelka, S. How Powerful are Graph Neural Networks? arXiv:1810.00826 [cs, stat] (2019).

    """

    def __init__(self,
                 num_features: int,
                 embedding_dim: int,
                 num_gc_layers: int = 5):
        dim = int(
            embedding_dim / num_gc_layers
        )  # the output dimension of this encoder is modified by the number of GC layers, so this is necessary to ensure that the output dimension is consistent with the InfoGraphEncoder
        super().__init__()

        self.num_gc_layers = num_gc_layers

        self.convs = torch.nn.ModuleList()
        self.bns = torch.nn.ModuleList()

        for i in range(num_gc_layers):
            if i == 0:
                nn = Sequential(Linear(num_features, dim), ReLU(),
                                Linear(dim, dim))
            else:
                nn = Sequential(Linear(dim, dim), ReLU(), Linear(dim, dim))
            conv = GINConv(nn)
            bn = torch.nn.BatchNorm1d(dim)

            self.convs.append(conv)
            self.bns.append(bn)

    def forward(self, data):
        """
        Encodes the input graph data.

        Parameters
        ----------
        data : BatchGraphData
            The batched input graph data.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            A tuple containing the encoded representation and intermediate embeddings.
        """
        xs = []
        x = data.node_features
        for i in range(self.num_gc_layers):

            x = F.relu(self.convs[i](x, data.edge_index))
            x = self.bns[i](x)
            xs.append(x)

        xpool = [global_add_pool(x, data.graph_index) for x in xs]
        x = torch.cat(xpool, 1)
        xs = torch.cat(xs, 1)
        return x, xs


class InfoGraphEncoder(torch.nn.Module):
    """
    The encoder for the InfoGraph model. It is a message passing graph convolutional
    network that produces encoded representations for molecular graph inputs.

    Parameters
    ----------
    num_features: int
        Number of node features for each input
    edge_features: int
        Number of edge features for each input
    dim: int
        Dimension of the embedding
    """

    def __init__(self, num_features, edge_features, dim):
        super().__init__()
        self.lin0 = torch.nn.Linear(num_features, dim)

        nn = Sequential(Linear(edge_features, 128), ReLU(),
                        Linear(128, dim * dim))
        self.conv = NNConv(dim, dim, nn, aggr='mean', root_weight=False)
        self.gru = GRU(dim, dim)

        self.set2set = Set2Set(dim, processing_steps=3)

    def forward(self, data):
        """
        Encode input graphs into an embedding and feature map.

        Parameters
        ----------
        data: Union[BatchGraphData, GraphData]
            Contains information about graphs.

        Returns
        -------
        torch.Tensor
            Encoded tensor of input data.
        torch.Tensor
            Feature map tensor of input data.

        """
        out = F.relu(self.lin0(data.node_features))
        h = out.unsqueeze(0)

        for i in range(3):
            m = F.relu(self.conv(out, data.edge_index, data.edge_features))
            out, h = self.gru(m.unsqueeze(0), h)
            out = out.squeeze(0)
            feat_map = out

        out = self.set2set(out, data.graph_index)
        return out, feat_map


class InfoGraphStar(torch.nn.Module):
    """
    The nn.Module for InfoGraphStar. This class defines the forward pass of InfoGraphStar.

    Parameters
    ----------
    encoder: torch.nn.Module
        The encoder for InfoGraphStar.
    unsup_encoder: torch.nn.Module
        The unsupervised encoder for InfoGraph, of identical architecture to encoder.
    ff1: torch.nn.Module
        The first feedforward layer for InfoGraphStar.
    ff2: torch.nn.Module
        The second feedforward layer for InfoGraphStar.
    fc1: torch.nn.Module
        The first fully connected layer for InfoGraphStar.
    fc2: torch.nn.Module
        The second fully connected layer for InfoGraphStar.

    References
    ----------
    .. [1] F.-Y. Sun, J. Hoffmann, V. Verma, and J. Tang, “InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Maximization.” arXiv, Jan. 17, 2020. http://arxiv.org/abs/1908.01000


    Example
    -------
    >>> import torch
    >>> import numpy as np
    >>> from deepchem.models.torch_models.infograph import InfoGraphStarModel
    >>> from deepchem.feat.molecule_featurizers import MolGraphConvFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> smiles = ['C1=CC=CC=C1', 'C1=CC=CC=C1C2=CC=CC=C2']
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> graphs = BatchGraphData(featurizer.featurize(smiles))
    >>> num_feat = 30
    >>> num_edge = 11
    >>> infographmodular = InfoGraphStarModel(num_feat,num_edge,64)
    >>> graphs.edge_features = torch.from_numpy(graphs.edge_features).to(infographmodular.device).float()
    >>> graphs.edge_index = torch.from_numpy(graphs.edge_index).to(infographmodular.device).long()
    >>> graphs.node_features = torch.from_numpy(graphs.node_features).to(infographmodular.device).float()
    >>> graphs.graph_index = torch.from_numpy(graphs.graph_index).to(infographmodular.device).long()
    >>> model = infographmodular.model
    >>> output = model(graphs).cpu().detach().numpy()

    """

    def __init__(self, encoder, unsup_encoder, ff1, ff2, fc1, fc2, local_d,
                 global_d):
        super().__init__()
        self.encoder = encoder
        self.unsup_encoder = unsup_encoder
        self.ff1 = ff1
        self.ff2 = ff2
        self.fc1 = fc1
        self.fc2 = fc2
        self.local_d = local_d
        self.global_d = global_d
        self.init_emb()

    def init_emb(self):
        """
        Initializes the parameters of the model by setting their values from a uniform distribution and filling the bias with 0.
        """
        for m in self.modules():
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.fill_(0.0)

    def forward(self, data):
        """
        Forward pass for InfoGraphStar.

        Parameters
        ----------
        data: Union[GraphData, BatchGraphData]
            The input data, either a single graph or a batch of graphs.
        """

        out, M = self.encoder(data)
        out = F.relu(self.fc1(out))
        pred = self.fc2(out)
        return pred


class InfoGraphStarModel(ModularTorchModel):
    """
    InfographStar is a semi-supervised graph convolutional network for predicting molecular properties.
    It aims to maximize the mutual information between the graph-level representation and the
    representations of substructures of different scales. It does this by producing graph-level
    encodings and substructure encodings, and then using a discriminator to classify if they
    are from the same molecule or not.

    Supervised training is done by using the graph-level encodings to predict the target property. Semi-supervised training is done by adding a loss term that maximizes the mutual information between the graph-level encodings and the substructure encodings to the supervised loss.
    These modes can be chosen by setting the training_mode parameter.

    To conduct training in unsupervised mode, use InfoGraphModel.

    References
    ----------
    .. [1] F.-Y. Sun, J. Hoffmann, V. Verma, and J. Tang, “InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Maximization.” arXiv, Jan. 17, 2020. http://arxiv.org/abs/1908.01000

    Parameters
    ----------
    num_features: int
        Number of node features for each input
    edge_features: int
        Number of edge features for each input
    embedding_dim: int
        Dimension of the embedding
    training_mode: str
        The mode to use for training. Options are 'supervised', 'semisupervised'. For unsupervised training, use InfoGraphModel.
    measure: str
        The divergence measure to use for the unsupervised loss. Options are 'GAN', 'JSD',
        'KL', 'RKL', 'X2', 'DV', 'H2', or 'W1'.
    average_loss: bool
        Whether to average the loss over the batch

    Examples
    --------
    >>> from deepchem.models.torch_models import InfoGraphStarModel
    >>> from deepchem.feat import MolGraphConvFeaturizer
    >>> from deepchem.data import NumpyDataset
    >>> import torch
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> X = featurizer.featurize(smiles)
    >>> y = torch.randint(0, 2, size=(2, 1)).float()
    >>> w = torch.ones(size=(2, 1)).float()
    >>> ds = NumpyDataset(X, y, w)
    >>> num_feat = max([ds.X[i].num_node_features for i in range(len(ds))])
    >>> edge_dim = max([ds.X[i].num_edge_features for i in range(len(ds))])
    >>> model = InfoGraphStarModel(num_feat, edge_dim, 15, training_mode='semisupervised')
    >>> loss = model.fit(ds, nb_epoch=1)
    """

    def __init__(self,
                 num_features,
                 edge_features,
                 embedding_dim,
                 task='supervised',
                 measure='JSD',
                 average_loss=True,
                 num_gc_layers=5,
                 **kwargs):

        self.edge_features = edge_features
        self.local = True
        self.prior = False
        self.gamma = .1
        self.num_features = num_features
        self.task = task
        if self.task == 'supervised':
            self.embedding_dim = embedding_dim
        elif self.task == 'semisupervised':
            self.embedding_dim = embedding_dim * num_gc_layers
            self.num_gc_layers = num_gc_layers
        self.localloss = LocalMutualInformationLoss()._create_pytorch_loss(
            measure, average_loss)
        self.globalloss = GlobalMutualInformationLoss()._create_pytorch_loss(
            measure, average_loss)

        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model, self.components, **kwargs)

    def build_components(self):
        """
        Builds the components of the InfoGraphStar model. InfoGraphStar works by maximizing the mutual information between the graph-level representation and the representations of substructures of different scales.

        It does this by producing graph-level encodings and substructure encodings, and then using a discriminator to classify if they are from the same molecule or not.

        The encoder is a graph convolutional network that produces the graph-level encodings and substructure encodings.

        In a supervised training mode, only 1 encoder is used and the encodings are not compared, while in a semi-supvervised training mode they are different in order to prevent negative transfer from the pretraining stage.

        The local discriminator is a multilayer perceptron that classifies if the substructure encodings are from the same molecule or not while the global discriminator classifies if the graph-level encodings are from the same molecule or not.

        Components list, type and description:
        --------------------------------------
        encoder: InfoGraphEncoder
        unsup_encoder: InfoGraphEncoder for supervised or GINEncoder for unsupervised training
        ff1: MultilayerPerceptron, feedforward network
        ff2: MultilayerPerceptron, feedforward network
        fc1: torch.nn.Linear, fully connected layer
        fc2: torch.nn.Linear, fully connected layer
        local_d: MultilayerPerceptron, local discriminator
        global_d: MultilayerPerceptron, global discriminator
        """
        if self.task == 'supervised':
            return {
                'encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'unsup_encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'ff1':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'ff2':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'fc1':
                    torch.nn.Linear(2 * self.embedding_dim, self.embedding_dim),
                'fc2':
                    torch.nn.Linear(self.embedding_dim, 1),
                'local_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True),
                'global_d':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True)
            }
        elif self.task == 'semisupervised':
            return {
                'encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'unsup_encoder':
                    GINEncoder(self.num_features, self.embedding_dim,
                               self.num_gc_layers),
                'ff1':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'ff2':
                    MultilayerPerceptron(self.embedding_dim, self.embedding_dim,
                                         (self.embedding_dim,)),
                'fc1':
                    torch.nn.Linear(2 * self.embedding_dim, self.embedding_dim),
                'fc2':
                    torch.nn.Linear(self.embedding_dim, 1),
                'local_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True),
                'global_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True)
            }

    def build_model(self):
        """
        Builds the InfoGraph model by unpacking the components dictionary and passing them to the InfoGraph nn.module.
        """
        return InfoGraphStar(**self.components)

    def loss_func(self, inputs, labels, weights):
        if self.task == 'semisupervised':
            sup_loss = F.mse_loss(self.model(inputs), labels)
            local_unsup_loss = self.local_unsup_loss(inputs)
            global_unsup_loss = self.global_unsup_loss(inputs, labels, weights)
            loss = sup_loss + local_unsup_loss + global_unsup_loss * self.learning_rate
            # original implementation also includes an option if not using a separate encoder:
            # loss = sup_loss + local_unsup_loss * self.learning_rate
            return (loss * weights).mean()
        else:
            sup_loss = F.mse_loss(self.model(inputs), labels)
            return (sup_loss * weights).mean()

    def local_unsup_loss(self, inputs):
        if self.task == 'semisupervised':
            y, M = self.components['unsup_encoder'](inputs)
        else:
            y, M = self.components['encoder'](inputs)
        g_enc = self.components['global_d'](y)
        l_enc = self.components['local_d'](M)

        if self.local:
            loss = self.localloss(l_enc, g_enc, inputs.graph_index)

        return loss

    def global_unsup_loss(self, inputs, labels, weights):
        y, M = self.components['encoder'](inputs)
        y_, M_ = self.components['unsup_encoder'](inputs)

        g_enc = self.components['ff1'](y)
        g_enc1 = self.components['ff2'](y_)

        loss = self.globalloss(g_enc, g_enc1)
        return loss

    def _prepare_batch(self, batch):
        """
        Prepares the batch for the model by converting the GraphData numpy arrays to torch tensors and moving them to the device.
        """
        inputs, labels, weights = batch
        inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)

        _, labels, weights = super()._prepare_batch(([], labels, weights))

        if (len(labels) != 0) and (len(weights) != 0):
            labels = labels[0]
            weights = weights[0]

        return inputs, labels, weights



import torch
import torch.nn as nn
from typing import List
from deepchem.models.torch_models.grover_layers import GroverTransEncoder


class GroverEmbedding(nn.Module):
    """GroverEmbedding layer.

    This layer is a simple wrapper over GroverTransEncoder layer for retrieving the embeddings from the GroverTransEncoder corresponding to the `embedding_output_type` chosen by the user.

    Parameters
    ----------
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    node_fdim: int
        the dimension of additional feature for node/atom.
    depth: int
        Dynamic message passing depth for use in MPNEncoder
    undirected: bool
        The message passing is undirected or not
    num_mt_block: int
        the number of message passing blocks.
    num_head: int
        the number of attention heads.
    embedding_output_type: str
        the type of output aggregation after message passing.
                                            atom_messages:      True                      False
        - "none": no aggregation         output size:     (num_atoms, hidden_size)    (num_bonds, hidden_size)
        -"atom":  aggregating to atom  output size:     (num_atoms, hidden_size)    (num_atoms, hidden_size)
        -"bond": aggragating to bond.   output size:     (num_bonds, hidden_size)    (num_bonds, hidden_size)
        -"both": aggregating to atom&bond. output size:  (num_atoms, hidden_size)    (num_bonds, hidden_size)
                                                        (num_bonds, hidden_size)    (num_atoms, hidden_size)
    """

    def __init__(self,
                 node_fdim,
                 edge_fdim,
                 embedding_output_type,
                 hidden_size=128,
                 depth=1,
                 undirected=False,
                 dropout=0.2,
                 activation='relu',
                 num_mt_block=1,
                 num_heads=4,
                 bias=False,
                 res_connection=False):
        super(GroverEmbedding, self).__init__()
        self.embedding_output_type = embedding_output_type
        self.encoders = GroverTransEncoder(
            hidden_size=hidden_size,
            edge_fdim=edge_fdim,
            node_fdim=node_fdim,
            depth=depth,
            undirected=undirected,
            dropout=dropout,
            activation=activation,
            num_mt_block=num_mt_block,
            num_heads=num_heads,
            embedding_output_type=embedding_output_type,
            bias=bias,
            res_connection=res_connection)

    def forward(self, graph_batch: List[torch.Tensor]):
        """Forward function

        Parameters
        ----------
        graph_batch: List[torch.Tensor]
            A list containing f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
        """
        output = self.encoders(graph_batch)
        if self.embedding_output_type == 'atom':
            return {
                "atom_from_atom": output[0],
                "atom_from_bond": output[1],
                "bond_from_atom": None,
                "bond_from_bond": None
            }  # atom_from_atom, atom_from_bond
        elif self.embedding_output_type == 'bond':
            return {
                "atom_from_atom": None,
                "atom_from_bond": None,
                "bond_from_atom": output[0],
                "bond_from_bond": output[1]
            }  # bond_from_atom, bond_from_bond
        elif self.embedding_output_type == "both":
            return {
                "atom_from_atom": output[0][0],
                "bond_from_atom": output[0][1],
                "atom_from_bond": output[1][0],
                "bond_from_bond": output[1][1]
            }



# flake8: noqa

import numpy as np
import deepchem as dc
from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.mol_graphs import ConvMol, WeaveMol
from deepchem.data import DiskDataset
import logging
from typing import Optional, List, Tuple, Union, Iterable
from deepchem.utils.typing import RDKitMol, RDKitAtom
from deepchem.utils.molecule_feature_utils import one_hot_encode


def one_of_k_encoding(x, allowable_set):
    """Encodes elements of a provided set as integers.

    Parameters
    ----------
    x: object
        Must be present in `allowable_set`.
    allowable_set: list
        List of allowable quantities.

    Example
    -------
    >>> import deepchem as dc
    >>> dc.feat.graph_features.one_of_k_encoding("a", ["a", "b", "c"])
    [True, False, False]

    Raises
    ------
    `ValueError` if `x` is not in `allowable_set`.
    """
    if x not in allowable_set:
        raise ValueError("input {0} not in allowable set{1}:".format(
            x, allowable_set))
    return list(map(lambda s: x == s, allowable_set))


def one_of_k_encoding_unk(x, allowable_set):
    """Maps inputs not in the allowable set to the last element.

    Unlike `one_of_k_encoding`, if `x` is not in `allowable_set`, this method
    pretends that `x` is the last element of `allowable_set`.

    Parameters
    ----------
    x: object
        Must be present in `allowable_set`.
    allowable_set: list
        List of allowable quantities.

    Examples
    --------
    >>> dc.feat.graph_features.one_of_k_encoding_unk("s", ["a", "b", "c"])
    [False, False, True]

    """
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: x == s, allowable_set))


def get_intervals(l):
    """For list of lists, gets the cumulative products of the lengths

    Note that we add 1 to the lengths of all lists (to avoid an empty list
    propagating a 0).

    Parameters
    ----------
    l: list of lists
        Returns the cumulative product of these lengths.

    Examples
    --------
    >>> dc.feat.graph_features.get_intervals([[1], [1, 2], [1, 2, 3]])
    [1, 3, 12]

    >>> dc.feat.graph_features.get_intervals([[1], [], [1, 2], [1, 2, 3]])
    [1, 1, 3, 12]

    """
    intervals = len(l) * [0]
    # Initalize with 1
    intervals[0] = 1
    for k in range(1, len(l)):
        intervals[k] = (len(l[k]) + 1) * intervals[k - 1]

    return intervals


def safe_index(l, e):
    """Gets the index of e in l, providing an index of len(l) if not found

    Parameters
    ----------
    l: list
        List of values
    e: object
        Object to check whether `e` is in `l`

    Examples
    --------
    >>> dc.feat.graph_features.safe_index([1, 2, 3], 1)
    0
    >>> dc.feat.graph_features.safe_index([1, 2, 3], 7)
    3

    """
    try:
        return l.index(e)
    except ValueError:
        return len(l)


class GraphConvConstants(object):
    """This class defines a collection of constants which are useful for graph convolutions on molecules."""
    possible_atom_list = [
        'C', 'N', 'O', 'S', 'F', 'P', 'Cl', 'Mg', 'Na', 'Br', 'Fe', 'Ca', 'Cu',
        'Mc', 'Pd', 'Pb', 'K', 'I', 'Al', 'Ni', 'Mn'
    ]
    """Allowed Numbers of Hydrogens"""
    possible_numH_list = [0, 1, 2, 3, 4]
    """Allowed Valences for Atoms"""
    possible_valence_list = [0, 1, 2, 3, 4, 5, 6]
    """Allowed Formal Charges for Atoms"""
    possible_formal_charge_list = [-3, -2, -1, 0, 1, 2, 3]
    """This is a placeholder for documentation. These will be replaced with corresponding values of the rdkit HybridizationType"""
    possible_hybridization_list = ["SP", "SP2", "SP3", "SP3D", "SP3D2"]
    """Allowed number of radical electrons."""
    possible_number_radical_e_list = [0, 1, 2]
    """Allowed types of Chirality"""
    possible_chirality_list = ['R', 'S']
    """The set of all values allowed."""
    reference_lists = [
        possible_atom_list, possible_numH_list, possible_valence_list,
        possible_formal_charge_list, possible_number_radical_e_list,
        possible_hybridization_list, possible_chirality_list
    ]
    """The number of different values that can be taken. See `get_intervals()`"""
    intervals = get_intervals(reference_lists)
    """Possible stereochemistry. We use E-Z notation for stereochemistry
https://en.wikipedia.org/wiki/E%E2%80%93Z_notation"""
    possible_bond_stereo = ["STEREONONE", "STEREOANY", "STEREOZ", "STEREOE"]
    """Number of different bond types not counting stereochemistry."""
    bond_fdim_base = 6


def get_feature_list(atom):
    """Returns a list of possible features for this atom.

    Parameters
    ----------
    atom: RDKit.Chem.rdchem.Atom
        Atom to get features for

    Examples
    --------
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles("C")
    >>> atom = mol.GetAtoms()[0]
    >>> features = dc.feat.graph_features.get_feature_list(atom)
    >>> type(features)
    <class 'list'>
    >>> len(features)
    6

    Note
    ----
    This method requires RDKit to be installed.

    Returns
    -------
    features: list
        List of length 6. The i-th value in this list provides the index of the
        atom in the corresponding feature value list. The 6 feature values lists
        for this function are `[GraphConvConstants.possible_atom_list,
        GraphConvConstants.possible_numH_list,
        GraphConvConstants.possible_valence_list,
        GraphConvConstants.possible_formal_charge_list,
        GraphConvConstants.possible_num_radical_e_list]`.
    """
    possible_atom_list = GraphConvConstants.possible_atom_list
    possible_numH_list = GraphConvConstants.possible_numH_list
    possible_valence_list = GraphConvConstants.possible_valence_list
    possible_formal_charge_list = GraphConvConstants.possible_formal_charge_list
    possible_number_radical_e_list = GraphConvConstants.possible_number_radical_e_list
    possible_hybridization_list = GraphConvConstants.possible_hybridization_list
    # Replace the hybridization
    from rdkit import Chem
    #global possible_hybridization_list
    possible_hybridization_list = [
        Chem.rdchem.HybridizationType.SP, Chem.rdchem.HybridizationType.SP2,
        Chem.rdchem.HybridizationType.SP3, Chem.rdchem.HybridizationType.SP3D,
        Chem.rdchem.HybridizationType.SP3D2
    ]
    features = 6 * [0]
    features[0] = safe_index(possible_atom_list, atom.GetSymbol())
    features[1] = safe_index(possible_numH_list, atom.GetTotalNumHs())
    features[2] = safe_index(possible_valence_list, atom.GetImplicitValence())
    features[3] = safe_index(possible_formal_charge_list,
                             atom.GetFormalCharge())
    features[4] = safe_index(possible_number_radical_e_list,
                             atom.GetNumRadicalElectrons())

    features[5] = safe_index(possible_hybridization_list,
                             atom.GetHybridization())
    return features


def features_to_id(features, intervals):
    """Convert list of features into index using spacings provided in intervals

    Parameters
    ----------
    features: list
        List of features as returned by `get_feature_list()`
    intervals: list
        List of intervals as returned by `get_intervals()`

    Returns
    -------
    id: int
        The index in a feature vector given by the given set of features.
    """
    id = 0
    for k in range(len(intervals)):
        id += features[k] * intervals[k]

    # Allow 0 index to correspond to null molecule 1
    id = id + 1
    return id


def id_to_features(id, intervals):
    """Given an index in a feature vector, return the original set of features.

    Parameters
    ----------
    id: int
        The index in a feature vector given by the given set of features.
    intervals: list
        List of intervals as returned by `get_intervals()`

    Returns
    -------
    features: list
        List of features as returned by `get_feature_list()`
    """
    features = 6 * [0]

    # Correct for null
    id -= 1

    for k in range(0, 6 - 1):
        # print(6-k-1, id)
        features[6 - k - 1] = id // intervals[6 - k - 1]
        id -= features[6 - k - 1] * intervals[6 - k - 1]
    # Correct for last one
    features[0] = id
    return features


def atom_to_id(atom):
    """Return a unique id corresponding to the atom type

    Parameters
    ----------
    atom: RDKit.Chem.rdchem.Atom
        Atom to convert to ids.

    Returns
    -------
    id: int
        The index in a feature vector given by the given set of features.
    """
    features = get_feature_list(atom)
    return features_to_id(features, intervals)


def atom_features(atom,
                  bool_id_feat=False,
                  explicit_H=False,
                  use_chirality=False):
    """Helper method used to compute per-atom feature vectors.

    Many different featurization methods compute per-atom features such as ConvMolFeaturizer, WeaveFeaturizer. This method computes such features.

    Parameters
    ----------
    atom: RDKit.Chem.rdchem.Atom
        Atom to compute features on.
    bool_id_feat: bool, optional
        Return an array of unique identifiers corresponding to atom type.
    explicit_H: bool, optional
        If true, model hydrogens explicitly
    use_chirality: bool, optional
        If true, use chirality information.

    Returns
    -------
    features: np.ndarray
        An array of per-atom features.

    Examples
    --------
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('CCC')
    >>> atom = mol.GetAtoms()[0]
    >>> features = dc.feat.graph_features.atom_features(atom)
    >>> type(features)
    <class 'numpy.ndarray'>
    >>> features.shape
    (75,)

    """
    if bool_id_feat:
        return np.array([atom_to_id(atom)])
    else:
        from rdkit import Chem
        results = one_of_k_encoding_unk(
          atom.GetSymbol(),
          [
            'C',
            'N',
            'O',
            'S',
            'F',
            'Si',
            'P',
            'Cl',
            'Br',
            'Mg',
            'Na',
            'Ca',
            'Fe',
            'As',
            'Al',
            'I',
            'B',
            'V',
            'K',
            'Tl',
            'Yb',
            'Sb',
            'Sn',
            'Ag',
            'Pd',
            'Co',
            'Se',
            'Ti',
            'Zn',
            'H',  # H?
            'Li',
            'Ge',
            'Cu',
            'Au',
            'Ni',
            'Cd',
            'In',
            'Mn',
            'Zr',
            'Cr',
            'Pt',
            'Hg',
            'Pb',
            'Unknown'
          ]) + one_of_k_encoding(atom.GetDegree(),
                                 [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]) + \
                  one_of_k_encoding_unk(atom.GetImplicitValence(), [0, 1, 2, 3, 4, 5, 6]) + \
                  [atom.GetFormalCharge(), atom.GetNumRadicalElectrons()] + \
                  one_of_k_encoding_unk(atom.GetHybridization(), [
                    Chem.rdchem.HybridizationType.SP, Chem.rdchem.HybridizationType.SP2,
                    Chem.rdchem.HybridizationType.SP3, Chem.rdchem.HybridizationType.
                                        SP3D, Chem.rdchem.HybridizationType.SP3D2
                  ]) + [atom.GetIsAromatic()]
        # In case of explicit hydrogen(QM8, QM9), avoid calling `GetTotalNumHs`
        if not explicit_H:
            results = results + one_of_k_encoding_unk(atom.GetTotalNumHs(),
                                                      [0, 1, 2, 3, 4])
        if use_chirality:
            try:
                results = results + one_of_k_encoding_unk(
                    atom.GetProp('_CIPCode'),
                    ['R', 'S']) + [atom.HasProp('_ChiralityPossible')]
            except:
                results = results + [False, False
                                    ] + [atom.HasProp('_ChiralityPossible')]

        return np.array(results)


def bond_features(bond, use_chirality=False, use_extended_chirality=False):
    """Helper method used to compute bond feature vectors.

    Many different featurization methods compute bond features
    such as WeaveFeaturizer. This method computes such features.

    Parameters
    ----------
    bond: rdkit.Chem.rdchem.Bond
        Bond to compute features on.
    use_chirality: bool, optional
        If true, use chirality information.
    use_extended_chirality: bool, optional
        If true, use chirality information with upto 6 different types.

    Note
    ----
    This method requires RDKit to be installed.

    Returns
    -------
    bond_feats: np.ndarray
        Array of bond features. This is a 1-D array of length 6 if `use_chirality`
        is `False` else of length 10 with chirality encoded.

    bond_feats: Sequence[Union[bool, int, float]]
        List of bond features returned if `use_extended_chirality` is `True`.

    Examples
    --------
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('CCC')
    >>> bond = mol.GetBonds()[0]
    >>> bond_features = dc.feat.graph_features.bond_features(bond)
    >>> type(bond_features)
    <class 'numpy.ndarray'>
    >>> bond_features.shape
    (6,)

    Note
    ----
    This method requires RDKit to be installed.

    """
    try:
        from rdkit import Chem
    except ModuleNotFoundError:
        raise ImportError("This method requires RDKit to be installed.")
    bt = bond.GetBondType()
    bond_feats = [
        bt == Chem.rdchem.BondType.SINGLE, bt == Chem.rdchem.BondType.DOUBLE,
        bt == Chem.rdchem.BondType.TRIPLE, bt == Chem.rdchem.BondType.AROMATIC,
        bond.GetIsConjugated(),
        bond.IsInRing()
    ]
    if use_chirality:
        bond_feats = bond_feats + one_of_k_encoding_unk(
            str(bond.GetStereo()), GraphConvConstants.possible_bond_stereo)

    if use_extended_chirality:
        stereo = one_hot_encode(int(bond.GetStereo()), list(range(6)), True)
        stereo = [int(feature) for feature in stereo]
        bond_feats = bond_feats + stereo
        return bond_feats

    return np.array(bond_feats)


def max_pair_distance_pairs(mol: RDKitMol,
                            max_pair_distance: Optional[int] = None
                           ) -> np.ndarray:
    """Helper method which finds atom pairs within max_pair_distance graph distance.

    This helper method is used to find atoms which are within max_pair_distance
    graph_distance of one another. This is done by using the fact that the
    powers of an adjacency matrix encode path connectivity information. In
    particular, if `adj` is the adjacency matrix, then `adj**k` has a nonzero
    value at `(i, j)` if and only if there exists a path of graph distance `k`
    between `i` and `j`. To find all atoms within `max_pair_distance` of each
    other, we can compute the adjacency matrix powers `[adj, adj**2,
    ...,adj**max_pair_distance]` and find pairs which are nonzero in any of
    these matrices. Since adjacency matrices and their powers are positive
    numbers, this is simply the nonzero elements of `adj + adj**2 + ... +
    adj**max_pair_distance`.

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol
        RDKit molecules
    max_pair_distance: Optional[int], (default None)
        This value can be a positive integer or None. This
        parameter determines the maximum graph distance at which pair
        features are computed. For example, if `max_pair_distance==2`,
        then pair features are computed only for atoms at most graph
        distance 2 apart. If `max_pair_distance` is `None`, all pairs are
        considered (effectively infinite `max_pair_distance`)

    Examples
    --------
    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('CCC')
    >>> features = dc.feat.graph_features.max_pair_distance_pairs(mol, 1)
    >>> type(features)
    <class 'numpy.ndarray'>
    >>> features.shape  # (2, num_pairs)
    (2, 7)

    Returns
    -------
    np.ndarray
        Of shape `(2, num_pairs)` where `num_pairs` is the total number of pairs
        within `max_pair_distance` of one another.
    """
    from rdkit import Chem
    from rdkit.Chem import rdmolops
    N = len(mol.GetAtoms())
    if (max_pair_distance is None or max_pair_distance >= N):
        max_distance = N
    elif max_pair_distance is not None and max_pair_distance <= 0:
        raise ValueError(
            "max_pair_distance must either be a positive integer or None")
    elif max_pair_distance is not None:
        max_distance = max_pair_distance
    adj = rdmolops.GetAdjacencyMatrix(mol)
    # Handle edge case of self-pairs (i, i)
    sum_adj = np.eye(N)
    for i in range(max_distance):
        # Increment by 1 since we don't want 0-indexing
        power = i + 1
        sum_adj += np.linalg.matrix_power(adj, power)
    nonzero_locs = np.where(sum_adj != 0)
    num_pairs = len(nonzero_locs[0])
    # This creates a matrix of shape (2, num_pairs)
    pair_edges = np.reshape(np.array(list(zip(nonzero_locs))), (2, num_pairs))
    return pair_edges


def pair_features(
        mol: RDKitMol,
        bond_features_map: dict,
        bond_adj_list: List,
        bt_len: int = 6,
        graph_distance: bool = True,
        max_pair_distance: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Helper method used to compute atom pair feature vectors.

    Many different featurization methods compute atom pair features
    such as WeaveFeaturizer. Note that atom pair features could be
    for pairs of atoms which aren't necessarily bonded to one
    another.

    Parameters
    ----------
    mol: RDKit Mol
        Molecule to compute features on.
    bond_features_map: dict
        Dictionary that maps pairs of atom ids (say `(2, 3)` for a bond between
        atoms 2 and 3) to the features for the bond between them.
    bond_adj_list: list of lists
        `bond_adj_list[i]` is a list of the atom indices that atom `i` shares a
        bond with . This list is symmetrical so if `j in bond_adj_list[i]` then `i
        in bond_adj_list[j]`.
    bt_len: int, optional (default 6)
        The number of different bond types to consider.
    graph_distance: bool, optional (default True)
        If true, use graph distance between molecules. Else use euclidean
        distance. The specified `mol` must have a conformer. Atomic
        positions will be retrieved by calling `mol.getConformer(0)`.
    max_pair_distance: Optional[int], (default None)
        This value can be a positive integer or None. This
        parameter determines the maximum graph distance at which pair
        features are computed. For example, if `max_pair_distance==2`,
        then pair features are computed only for atoms at most graph
        distance 2 apart. If `max_pair_distance` is `None`, all pairs are
        considered (effectively infinite `max_pair_distance`)

    Note
    ----
    This method requires RDKit to be installed.

    Returns
    -------
    features: np.ndarray
        Of shape `(N_edges, bt_len + max_distance + 1)`. This is the array
        of pairwise features for all atom pairs, where N_edges is the
        number of edges within max_pair_distance of one another in this
        molecules.
    pair_edges: np.ndarray
        Of shape `(2, num_pairs)` where `num_pairs` is the total number of
        pairs within `max_pair_distance` of one another.
    """
    if graph_distance:
        max_distance = 7
    else:
        max_distance = 1
    N = mol.GetNumAtoms()
    pair_edges = max_pair_distance_pairs(mol, max_pair_distance)
    num_pairs = pair_edges.shape[1]
    N_edges = pair_edges.shape[1]
    features = np.zeros((N_edges, bt_len + max_distance + 1))
    # Get mapping
    mapping = {}
    for n in range(N_edges):
        a1, a2 = pair_edges[:, n]
        mapping[(int(a1), int(a2))] = n
    num_atoms = mol.GetNumAtoms()
    rings = mol.GetRingInfo().AtomRings()
    for a1 in range(num_atoms):
        for a2 in bond_adj_list[a1]:
            # first `bt_len` features are bond features(if applicable)
            if (int(a1), int(a2)) not in mapping:
                raise ValueError(
                    "Malformed molecule with bonds not in specified graph distance."
                )
            else:
                n = mapping[(int(a1), int(a2))]
            features[n, :bt_len] = np.asarray(bond_features_map[tuple(
                sorted((a1, a2)))],
                                              dtype=float)
        for ring in rings:
            if a1 in ring:
                for a2 in ring:
                    if (int(a1), int(a2)) not in mapping:
                        # For ring pairs outside max pairs distance continue
                        continue
                    else:
                        n = mapping[(int(a1), int(a2))]
                    # `bt_len`-th feature is if the pair of atoms are in the same ring
                    if a2 == a1:
                        features[n, bt_len] = 0
                    else:
                        features[n, bt_len] = 1
        # graph distance between two atoms
        if graph_distance:
            # distance is a matrix of 1-hot encoded distances for all atoms
            distance = find_distance(a1,
                                     num_atoms,
                                     bond_adj_list,
                                     max_distance=max_distance)
            for a2 in range(num_atoms):
                if (int(a1), int(a2)) not in mapping:
                    # For ring pairs outside max pairs distance continue
                    continue
                else:
                    n = mapping[(int(a1), int(a2))]
                    features[n, bt_len + 1:] = distance[a2]
    # Euclidean distance between atoms
    if not graph_distance:
        coords = np.zeros((N, 3))
        for atom in range(N):
            pos = mol.GetConformer(0).GetAtomPosition(atom)
            coords[atom, :] = [pos.x, pos.y, pos.z]
        features[:, :, -1] = np.sqrt(np.sum(np.square(
          np.stack([coords] * N, axis=1) - \
          np.stack([coords] * N, axis=0)), axis=2))

    return features, pair_edges


def find_distance(a1: RDKitAtom,
                  num_atoms: int,
                  bond_adj_list,
                  max_distance=7) -> np.ndarray:
    """Computes distances from provided atom.

    Parameters
    ----------
    a1: RDKit atom
        The source atom to compute distances from.
    num_atoms: int
        The total number of atoms.
    bond_adj_list: list of lists
        `bond_adj_list[i]` is a list of the atom indices that atom `i` shares a
        bond with. This list is symmetrical so if `j in bond_adj_list[i]` then `i in
        bond_adj_list[j]`.
    max_distance: int, optional (default 7)
        The max distance to search.

    Returns
    -------
    distances: np.ndarray
        Of shape `(num_atoms, max_distance)`. Provides a one-hot encoding of the
        distances. That is, `distances[i]` is a one-hot encoding of the distance
        from `a1` to atom `i`.
    """
    distance = np.zeros((num_atoms, max_distance))
    radial = 0
    # atoms `radial` bonds away from `a1`
    adj_list = set(bond_adj_list[a1])
    # atoms less than `radial` bonds away
    all_list = set([a1])
    while radial < max_distance:
        distance[list(adj_list), radial] = 1
        all_list.update(adj_list)
        # find atoms `radial`+1 bonds away
        next_adj = set()
        for adj in adj_list:
            next_adj.update(bond_adj_list[adj])
        adj_list = next_adj - all_list
        radial = radial + 1
    return distance


class ConvMolFeaturizer(MolecularFeaturizer):
    """This class implements the featurization to implement Duvenaud graph convolutions.

    Duvenaud graph convolutions [1]_ construct a vector of descriptors for each
    atom in a molecule. The featurizer computes that vector of local descriptors.

    Examples
    ---------
    >>> import deepchem as dc
    >>> smiles = ["C", "CCC"]
    >>> featurizer=dc.feat.ConvMolFeaturizer(per_atom_fragmentation=False)
    >>> f = featurizer.featurize(smiles)
    >>> # Using ConvMolFeaturizer to create featurized fragments derived from molecules of interest.
    ... # This is used only in the context of performing interpretation of models using atomic
    ... # contributions (atom-based model interpretation)
    ... smiles = ["C", "CCC"]
    >>> featurizer=dc.feat.ConvMolFeaturizer(per_atom_fragmentation=True)
    >>> f = featurizer.featurize(smiles)
    >>> len(f) # contains 2 lists with  featurized fragments from 2 mols
    2

    See Also
    --------
    Detailed examples of `GraphConvModel` interpretation are provided in Tutorial #28

    References
    ---------

    .. [1] Duvenaud, David K., et al. "Convolutional networks on graphs for
        learning molecular fingerprints." Advances in neural information
        processing systems. 2015.

    Note
    ----
    This class requires RDKit to be installed.
    """
    name = ['conv_mol']

    def __init__(self,
                 master_atom: bool = False,
                 use_chirality: bool = False,
                 atom_properties: Iterable[str] = [],
                 per_atom_fragmentation: bool = False):
        """
        Parameters
        ----------
        master_atom: Boolean
            if true create a fake atom with bonds to every other atom.
            the initialization is the mean of the other atom features in
            the molecule.  This technique is briefly discussed in
            Neural Message Passing for Quantum Chemistry
            https://arxiv.org/pdf/1704.01212.pdf
        use_chirality: Boolean
            if true then make the resulting atom features aware of the
            chirality of the molecules in question
        atom_properties: list of string or None
            properties in the RDKit Mol object to use as additional
            atom-level features in the larger molecular feature.  If None,
            then no atom-level properties are used.  Properties should be in the
            RDKit mol object should be in the form
            atom XXXXXXXX NAME
            where XXXXXXXX is a zero-padded 8 digit number coresponding to the
            zero-indexed atom index of each atom and NAME is the name of the property
            provided in atom_properties.  So "atom 00000000 sasa" would be the
            name of the molecule level property in mol where the solvent
            accessible surface area of atom 0 would be stored.
        per_atom_fragmentation: Boolean
            If True, then multiple "atom-depleted" versions of each molecule will be created (using featurize() method).
            For each molecule, atoms are removed one at a time and the resulting molecule is featurized.
            The result is a list of ConvMol objects,
            one with each heavy atom removed. This is useful for subsequent model interpretation: finding atoms
            favorable/unfavorable for (modelled) activity. This option is typically used in combination
            with a FlatteningTransformer to split the lists into separate samples.

            Since ConvMol is an object and not a numpy array, need to set dtype to
            object.
        """
        self.dtype = object
        self.master_atom = master_atom
        self.use_chirality = use_chirality
        self.atom_properties = list(atom_properties)
        self.per_atom_fragmentation = per_atom_fragmentation

    def featurize(self,
                  datapoints: Union[RDKitMol, str, Iterable[RDKitMol],
                                    Iterable[str]],
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """
        Override parent: aim is to add handling atom-depleted molecules featurization

        Parameters
        ----------
        datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
            RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
            strings.
        log_every_n: int, default 1000
            Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
            A numpy array containing a featurized representation of `datapoints`.
        """
        if 'molecules' in kwargs and datapoints is None:
            datapoints = kwargs.get("molecules")
            raise DeprecationWarning(
                'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
            )

        features = super(ConvMolFeaturizer, self).featurize(datapoints,
                                                            log_every_n=1000)
        if self.per_atom_fragmentation:
            # create temporary valid ids serving to filter out failed featurizations from every sublist
            # of features (i.e. every molecules' frags list), and also totally failed sublists.
            # This makes output digestable by Loaders
            valid_frag_inds = [[
                True if np.array(elt).size > 0 else False for elt in f
            ] for f in features]
            features = np.array(
                [[elt for (is_valid, elt) in zip(l, m) if is_valid
                 ] for (l, m) in zip(valid_frag_inds, features) if any(l)],
                dtype=object)
        return features

    def _get_atom_properties(self, atom):
        """
        For a given input RDKit atom return the values of the properties
        requested when initializing the featurize.  See the __init__ of the
        class for a full description of the names of the properties

        Parameters
        ----------
        atom: RDKit.rdchem.Atom
            Atom to get the properties of
        returns a numpy lists of floats of the same size as self.atom_properties
        """
        values = []
        for prop in self.atom_properties:
            mol_prop_name = str("atom %08d %s" % (atom.GetIdx(), prop))
            try:
                values.append(float(atom.GetOwningMol().GetProp(mol_prop_name)))
            except KeyError:
                raise KeyError("No property %s found in %s in %s" %
                               (mol_prop_name, atom.GetOwningMol(), self))
        return np.array(values)

    def _featurize(self, mol):
        """Encodes mol as a ConvMol object.
        If per_atom_fragmentation is True,
        then for each molecule a list of ConvMolObjects
        will be created"""

        def per_atom(n, a):
            """
            Enumerates fragments resulting from mol object,
            s.t. each fragment = mol with single atom removed (all possible removals are enumerated)
            Goes over nodes, deletes one at a time and updates adjacency list of lists (removes connections to that node)

            Parameters
            ----------
            n: np.array of nodes (number_of_nodes X number_of_features)
            a: list of nested lists of adjacent node pairs

            """
            for i in range(n.shape[0]):
                new_n = np.delete(n, (i), axis=0)
                new_a = []
                for j, node_pair in enumerate(a):
                    if i != j:  # don't need this pair, no more connections to deleted node
                        tmp_node_pair = []
                        for v in node_pair:
                            if v < i:
                                tmp_node_pair.append(v)
                            elif v > i:
                                tmp_node_pair.append(
                                    v - 1
                                )  # renumber node, because of offset after node deletion
                        new_a.append(tmp_node_pair)
                yield new_n, new_a

        # Get the node features
        idx_nodes = [(a.GetIdx(),
                      np.concatenate(
                          (atom_features(a, use_chirality=self.use_chirality),
                           self._get_atom_properties(a))))
                     for a in mol.GetAtoms()]

        idx_nodes.sort()  # Sort by ind to ensure same order as rd_kit
        idx, nodes = list(zip(*idx_nodes))

        # Stack nodes into an array
        nodes = np.vstack(nodes)
        if self.master_atom:
            master_atom_features = np.expand_dims(np.mean(nodes, axis=0),
                                                  axis=0)
            nodes = np.concatenate([nodes, master_atom_features], axis=0)

        # Get bond lists with reverse edges included
        edge_list = [
            (b.GetBeginAtomIdx(), b.GetEndAtomIdx()) for b in mol.GetBonds()
        ]
        # Get canonical adjacency list
        canon_adj_list = [[] for mol_id in range(len(nodes))]
        for edge in edge_list:
            canon_adj_list[edge[0]].append(edge[1])
            canon_adj_list[edge[1]].append(edge[0])

        if self.master_atom:
            fake_atom_index = len(nodes) - 1
            for index in range(len(nodes) - 1):
                canon_adj_list[index].append(fake_atom_index)

        if not self.per_atom_fragmentation:
            return ConvMol(nodes, canon_adj_list)
        else:
            return [ConvMol(n, a) for n, a in per_atom(nodes, canon_adj_list)]

    def feature_length(self):
        return 75 + len(self.atom_properties)

    def __hash__(self):
        atom_properties = tuple(self.atom_properties)
        return hash((self.master_atom, self.use_chirality, atom_properties))

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return self.master_atom == other.master_atom and \
               self.use_chirality == other.use_chirality and \
               tuple(self.atom_properties) == tuple(other.atom_properties)


class WeaveFeaturizer(MolecularFeaturizer):
    """This class implements the featurization to implement Weave convolutions.

    Weave convolutions were introduced in [1]_. Unlike Duvenaud graph
    convolutions, weave convolutions require a quadratic matrix of interaction
    descriptors for each pair of atoms. These extra descriptors may provide for
    additional descriptive power but at the cost of a larger featurized dataset.


    Examples
    --------
    >>> import deepchem as dc
    >>> mols = ["CCC"]
    >>> featurizer = dc.feat.WeaveFeaturizer()
    >>> features = featurizer.featurize(mols)
    >>> type(features[0])
    <class 'deepchem.feat.mol_graphs.WeaveMol'>
    >>> features[0].get_num_atoms() # 3 atoms in compound
    3
    >>> features[0].get_num_features() # feature size
    75
    >>> type(features[0].get_atom_features())
    <class 'numpy.ndarray'>
    >>> features[0].get_atom_features().shape
    (3, 75)
    >>> type(features[0].get_pair_features())
    <class 'numpy.ndarray'>
    >>> features[0].get_pair_features().shape
    (9, 14)


    References
    ----------
    .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond
        fingerprints." Journal of computer-aided molecular design 30.8 (2016):
        595-608.

    Note
    ----
    This class requires RDKit to be installed.
    """

    name = ['weave_mol']

    def __init__(self,
                 graph_distance: bool = True,
                 explicit_H: bool = False,
                 use_chirality: bool = False,
                 max_pair_distance: Optional[int] = None):
        """Initialize this featurizer with set parameters.

        Parameters
        ----------
        graph_distance: bool, (default True)
            If True, use graph distance for distance features. Otherwise, use
            Euclidean distance. Note that this means that molecules that this
            featurizer is invoked on must have valid conformer information if this
            option is set.
        explicit_H: bool, (default False)
            If true, model hydrogens in the molecule.
        use_chirality: bool, (default False)
            If true, use chiral information in the featurization
        max_pair_distance: Optional[int], (default None)
            This value can be a positive integer or None. This
            parameter determines the maximum graph distance at which pair
            features are computed. For example, if `max_pair_distance==2`,
            then pair features are computed only for atoms at most graph
            distance 2 apart. If `max_pair_distance` is `None`, all pairs are
            considered (effectively infinite `max_pair_distance`)
        """
        # Distance is either graph distance(True) or Euclidean distance(False,
        # only support datasets providing Cartesian coordinates)
        self.graph_distance = graph_distance
        # Set dtype
        self.dtype = object
        # If includes explicit hydrogens
        self.explicit_H = explicit_H
        # If uses use_chirality
        self.use_chirality = use_chirality
        if isinstance(max_pair_distance, int) and max_pair_distance <= 0:
            raise ValueError(
                "max_pair_distance must either be a positive integer or None")
        self.max_pair_distance = max_pair_distance
        if self.use_chirality:
            self.bt_len = int(GraphConvConstants.bond_fdim_base) + len(
                GraphConvConstants.possible_bond_stereo)
        else:
            self.bt_len = int(GraphConvConstants.bond_fdim_base)

    def _featurize(self, mol):
        """Encodes mol as a WeaveMol object."""
        # Atom features
        idx_nodes = [(a.GetIdx(),
                      atom_features(a,
                                    explicit_H=self.explicit_H,
                                    use_chirality=self.use_chirality))
                     for a in mol.GetAtoms()]
        idx_nodes.sort()  # Sort by ind to ensure same order as rd_kit
        idx, nodes = list(zip(*idx_nodes))

        # Stack nodes into an array
        nodes = np.vstack(nodes)

        # Get bond lists
        bond_features_map = {}
        for b in mol.GetBonds():
            bond_features_map[tuple(
                sorted([b.GetBeginAtomIdx(),
                        b.GetEndAtomIdx()
                       ]))] = bond_features(b, use_chirality=self.use_chirality)

        # Get canonical adjacency list
        bond_adj_list = [[] for mol_id in range(len(nodes))]
        for bond in bond_features_map.keys():
            bond_adj_list[bond[0]].append(bond[1])
            bond_adj_list[bond[1]].append(bond[0])

        # Calculate pair features
        pairs, pair_edges = pair_features(
            mol,
            bond_features_map,
            bond_adj_list,
            bt_len=self.bt_len,
            graph_distance=self.graph_distance,
            max_pair_distance=self.max_pair_distance)

        return WeaveMol(nodes, pairs, pair_edges)



"""Sequence to sequence translation models."""

from deepchem.models import KerasModel, layers
from heapq import heappush, heappushpop
import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Input, Layer, Dense, Dropout, GRU, Lambda, Conv1D, Flatten, BatchNormalization


class VariationalRandomizer(Layer):
    """Add random noise to the embedding and include a corresponding loss."""

    def __init__(self, embedding_dimension, annealing_start_step,
                 annealing_final_step, **kwargs):
        super(VariationalRandomizer, self).__init__(**kwargs)
        self._embedding_dimension = embedding_dimension
        self._annealing_final_step = annealing_final_step
        self._annealing_start_step = annealing_start_step
        self.dense_mean = Dense(embedding_dimension)
        self.dense_stddev = Dense(embedding_dimension)
        self.combine = layers.CombineMeanStd(training_only=True)

    def call(self, inputs, training=True):
        input, global_step = inputs
        embedding_mean = self.dense_mean(input)
        embedding_stddev = self.dense_stddev(input)
        embedding = self.combine([embedding_mean, embedding_stddev],
                                 training=training)
        mean_sq = embedding_mean * embedding_mean
        stddev_sq = embedding_stddev * embedding_stddev
        kl = mean_sq + stddev_sq - tf.math.log(stddev_sq + 1e-20) - 1
        anneal_steps = self._annealing_final_step - self._annealing_start_step
        if anneal_steps > 0:
            current_step = tf.cast(global_step,
                                   tf.float32) - self._annealing_start_step
            anneal_frac = tf.maximum(0.0, current_step) / anneal_steps
            kl_scale = tf.minimum(1.0, anneal_frac * anneal_frac)
        else:
            kl_scale = 1.0
        self.add_loss(0.5 * kl_scale * tf.reduce_mean(kl))
        return embedding


class SeqToSeq(KerasModel):
    """Implements sequence to sequence translation models.

    The model is based on the description in Sutskever et al., "Sequence to
    Sequence Learning with Neural Networks" (https://arxiv.org/abs/1409.3215),
    although this implementation uses GRUs instead of LSTMs.  The goal is to
    take sequences of tokens as input, and translate each one into a different
    output sequence.  The input and output sequences can both be of variable
    length, and an output sequence need not have the same length as the input
    sequence it was generated from.  For example, these models were originally
    developed for use in natural language processing.  In that context, the
    input might be a sequence of English words, and the output might be a
    sequence of French words.  The goal would be to train the model to translate
    sentences from English to French.

    The model consists of two parts called the "encoder" and "decoder".  Each one
    consists of a stack of recurrent layers.  The job of the encoder is to
    transform the input sequence into a single, fixed length vector called the
    "embedding".  That vector contains all relevant information from the input
    sequence.  The decoder then transforms the embedding vector into the output
    sequence.

    These models can be used for various purposes.  First and most obviously,
    they can be used for sequence to sequence translation.  In any case where you
    have sequences of tokens, and you want to translate each one into a different
    sequence, a SeqToSeq model can be trained to perform the translation.

    Another possible use case is transforming variable length sequences into
    fixed length vectors.  Many types of models require their inputs to have a
    fixed shape, which makes it difficult to use them with variable sized inputs
    (for example, when the input is a molecule, and different molecules have
    different numbers of atoms).  In that case, you can train a SeqToSeq model as
    an autoencoder, so that it tries to make the output sequence identical to the
    input one.  That forces the embedding vector to contain all information from
    the original sequence.  You can then use the encoder for transforming
    sequences into fixed length embedding vectors, suitable to use as inputs to
    other types of models.

    Another use case is to train the decoder for use as a generative model.  Here
    again you begin by training the SeqToSeq model as an autoencoder.  Once
    training is complete, you can supply arbitrary embedding vectors, and
    transform each one into an output sequence.  When used in this way, you
    typically train it as a variational autoencoder.  This adds random noise to
    the encoder, and also adds a constraint term to the loss that forces the
    embedding vector to have a unit Gaussian distribution.  You can then pick
    random vectors from a Gaussian distribution, and the output sequences should
    follow the same distribution as the training data.

    When training as a variational autoencoder, it is best to use KL cost
    annealing, as described in https://arxiv.org/abs/1511.06349.  The constraint
    term in the loss is initially set to 0, so the optimizer just tries to
    minimize the reconstruction loss.  Once it has made reasonable progress
    toward that, the constraint term can be gradually turned back on.  The range
    of steps over which this happens is configurable.
    """

    sequence_end = object()

    def __init__(self,
                 input_tokens,
                 output_tokens,
                 max_output_length,
                 encoder_layers=4,
                 decoder_layers=4,
                 embedding_dimension=512,
                 dropout=0.0,
                 reverse_input=True,
                 variational=False,
                 annealing_start_step=5000,
                 annealing_final_step=10000,
                 **kwargs):
        """Construct a SeqToSeq model.

        In addition to the following arguments, this class also accepts all the keyword arguments
        from TensorGraph.

        Parameters
        ----------
        input_tokens: list
            a list of all tokens that may appear in input sequences
        output_tokens: list
            a list of all tokens that may appear in output sequences
        max_output_length: int
            the maximum length of output sequence that may be generated
        encoder_layers: int
            the number of recurrent layers in the encoder
        decoder_layers: int
            the number of recurrent layers in the decoder
        embedding_dimension: int
            the width of the embedding vector.  This also is the width of all
            recurrent layers.
        dropout: float
            the dropout probability to use during training
        reverse_input: bool
            if True, reverse the order of input sequences before sending them into
            the encoder.  This can improve performance when working with long sequences.
        variational: bool
            if True, train the model as a variational autoencoder.  This adds random
            noise to the encoder, and also constrains the embedding to follow a unit
            Gaussian distribution.
        annealing_start_step: int
            the step (that is, batch) at which to begin turning on the constraint term
            for KL cost annealing
        annealing_final_step: int
            the step (that is, batch) at which to finish turning on the constraint term
            for KL cost annealing
        """
        if SeqToSeq.sequence_end not in input_tokens:
            input_tokens = input_tokens + [SeqToSeq.sequence_end]
        if SeqToSeq.sequence_end not in output_tokens:
            output_tokens = output_tokens + [SeqToSeq.sequence_end]
        self._input_tokens = input_tokens
        self._output_tokens = output_tokens
        self._input_dict = dict((x, i) for i, x in enumerate(input_tokens))
        self._output_dict = dict((x, i) for i, x in enumerate(output_tokens))
        self._max_output_length = max_output_length
        self._embedding_dimension = embedding_dimension
        self._reverse_input = reverse_input
        self.encoder = self._create_encoder(encoder_layers, dropout)
        self.decoder = self._create_decoder(decoder_layers, dropout)
        features = self._create_features()
        gather_indices = Input(shape=(2,), dtype=tf.int32)
        global_step = Input(shape=tuple(), dtype=tf.int32)
        embedding = self.encoder([features, gather_indices])
        self._embedding = self.encoder([features, gather_indices],
                                       training=False)
        if variational:
            randomizer = VariationalRandomizer(self._embedding_dimension,
                                               annealing_start_step,
                                               annealing_final_step)
            embedding = randomizer([self._embedding, global_step])
            self._embedding = randomizer([self._embedding, global_step],
                                         training=False)
        output = self.decoder(embedding)
        model = tf.keras.Model(inputs=[features, gather_indices, global_step],
                               outputs=output)
        super(SeqToSeq, self).__init__(model, self._create_loss(), **kwargs)

    def _create_features(self):
        return Input(shape=(None, len(self._input_tokens)))

    def _create_encoder(self, n_layers, dropout):
        """Create the encoder as a tf.keras.Model."""
        input = self._create_features()
        gather_indices = Input(shape=(2,), dtype=tf.int32)
        prev_layer = input
        for i in range(n_layers):
            if dropout > 0.0:
                prev_layer = Dropout(rate=dropout)(prev_layer)
            prev_layer = GRU(self._embedding_dimension,
                             return_sequences=True)(prev_layer)
        prev_layer = Lambda(lambda x: tf.gather_nd(x[0], x[1]))(
            [prev_layer, gather_indices])
        return tf.keras.Model(inputs=[input, gather_indices],
                              outputs=prev_layer)

    def _create_decoder(self, n_layers, dropout):
        """Create the decoder as a tf.keras.Model."""
        input = Input(shape=(self._embedding_dimension,))
        prev_layer = layers.Stack()(self._max_output_length * [input])
        for i in range(n_layers):
            if dropout > 0.0:
                prev_layer = Dropout(dropout)(prev_layer)
            prev_layer = GRU(self._embedding_dimension,
                             return_sequences=True)(prev_layer)
        output = Dense(len(self._output_tokens),
                       activation=tf.nn.softmax)(prev_layer)
        return tf.keras.Model(inputs=input, outputs=output)

    def _create_loss(self):
        """Create the loss function."""

        def loss_fn(outputs, labels, weights):
            prob = tf.reduce_sum(outputs[0] * labels[0], axis=2)
            mask = tf.reduce_sum(labels[0], axis=2)
            log_prob = tf.math.log(prob + 1e-20) * mask
            loss = -tf.reduce_mean(tf.reduce_sum(log_prob, axis=1))
            return loss + sum(self.model.losses)

        return loss_fn

    def fit_sequences(self,
                      sequences,
                      max_checkpoints_to_keep=5,
                      checkpoint_interval=1000,
                      restore=False):
        """Train this model on a set of sequences

        Parameters
        ----------
        sequences: iterable
            the training samples to fit to.  Each sample should be
            represented as a tuple of the form (input_sequence, output_sequence).
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        """
        self.fit_generator(self._generate_batches(sequences),
                           max_checkpoints_to_keep=max_checkpoints_to_keep,
                           checkpoint_interval=checkpoint_interval,
                           restore=restore)

    def predict_from_sequences(self, sequences, beam_width=5):
        """Given a set of input sequences, predict the output sequences.

        The prediction is done using a beam search with length normalization.

        Parameters
        ----------
        sequences: iterable
            the input sequences to generate a prediction for
        beam_width: int
            the beam width to use for searching.  Set to 1 to use a simple greedy search.
        """
        result = []
        for batch in self._batch_elements(sequences):
            features = self._create_input_array(batch)
            indices = np.array([(i, len(batch[i]) if i < len(batch) else 0)
                                for i in range(self.batch_size)])
            probs = self.predict_on_generator([[
                (features, indices, np.array(self.get_global_step())), None,
                None
            ]])
            for i in range(len(batch)):
                result.append(self._beam_search(probs[i], beam_width))
        return result

    def predict_from_embeddings(self, embeddings, beam_width=5):
        """Given a set of embedding vectors, predict the output sequences.

        The prediction is done using a beam search with length normalization.

        Parameters
        ----------
        embeddings: iterable
            the embedding vectors to generate predictions for
        beam_width: int
            the beam width to use for searching.  Set to 1 to use a simple greedy search.
        """
        result = []
        for batch in self._batch_elements(embeddings):
            embedding_array = np.zeros(
                (self.batch_size, self._embedding_dimension), dtype=np.float32)
            for i, e in enumerate(batch):
                embedding_array[i] = e
            probs = self.decoder(embedding_array, training=False)
            probs = probs.numpy()
            for i in range(len(batch)):
                result.append(self._beam_search(probs[i], beam_width))
        return result

    def predict_embeddings(self, sequences):
        """Given a set of input sequences, compute the embedding vectors.

        Parameters
        ----------
        sequences: iterable
            the input sequences to generate an embedding vector for
        """
        result = []
        for batch in self._batch_elements(sequences):
            features = self._create_input_array(batch)
            indices = np.array([(i, len(batch[i]) if i < len(batch) else 0)
                                for i in range(self.batch_size)])
            embeddings = self.predict_on_generator(
                [[(features, indices, np.array(self.get_global_step())), None,
                  None]],
                outputs=self._embedding)
            for i in range(len(batch)):
                result.append(embeddings[i])
        return np.array(result, dtype=np.float32)

    def _beam_search(self, probs, beam_width):
        """Perform a beam search for the most likely output sequence."""
        if beam_width == 1:
            # Do a simple greedy search.

            s = []
            for i in range(len(probs)):
                token = self._output_tokens[np.argmax(probs[i])]
                if token == SeqToSeq.sequence_end:
                    break
                s.append(token)
            return s

        # Do a beam search with length normalization.

        logprobs = np.log(probs)
        # Represent each candidate as (normalized prob, raw prob, sequence)
        candidates = [(0.0, 0.0, [])]
        for i in range(len(logprobs)):
            new_candidates = []
            for c in candidates:
                if len(c[2]) > 0 and c[2][-1] == SeqToSeq.sequence_end:
                    # This candidate sequence has already been terminated
                    if len(new_candidates) < beam_width:
                        heappush(new_candidates, c)
                    else:
                        heappushpop(new_candidates, c)
                else:
                    # Consider all possible tokens we could add to this candidate sequence.
                    for j, logprob in enumerate(logprobs[i]):
                        new_logprob = logprob + c[1]
                        newc = (new_logprob / (len(c[2]) + 1), new_logprob,
                                c[2] + [self._output_tokens[j]])
                        if len(new_candidates) < beam_width:
                            heappush(new_candidates, newc)
                        else:
                            heappushpop(new_candidates, newc)
            candidates = new_candidates
        return sorted(candidates)[-1][2][:-1]

    def _create_input_array(self, sequences):
        """Create the array describing the input sequences for a batch."""
        lengths = [len(x) for x in sequences]
        if self._reverse_input:
            sequences = [reversed(s) for s in sequences]
        features = np.zeros(
            (self.batch_size, max(lengths) + 1, len(self._input_tokens)),
            dtype=np.float32)
        for i, sequence in enumerate(sequences):
            for j, token in enumerate(sequence):
                features[i, j, self._input_dict[token]] = 1
        features[np.arange(len(sequences)), lengths,
                 self._input_dict[SeqToSeq.sequence_end]] = 1
        return features

    def _create_output_array(self, sequences):
        """Create the array describing the target sequences for a batch."""
        lengths = [len(x) for x in sequences]
        labels = np.zeros(
            (self.batch_size, self._max_output_length, len(
                self._output_tokens)),
            dtype=np.float32)
        end_marker_index = self._output_dict[SeqToSeq.sequence_end]
        for i, sequence in enumerate(sequences):
            for j, token in enumerate(sequence):
                labels[i, j, self._output_dict[token]] = 1
            for j in range(lengths[i], self._max_output_length):
                labels[i, j, end_marker_index] = 1
        return labels

    def _batch_elements(self, elements):
        """Combine elements into batches."""
        batch = []
        for s in elements:
            batch.append(s)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0:
            yield batch

    def _generate_batches(self, sequences):
        """Create feed_dicts for fitting."""
        for batch in self._batch_elements(sequences):
            inputs = []
            outputs = []
            for input, output in batch:
                inputs.append(input)
                outputs.append(output)
            for i in range(len(inputs), self.batch_size):
                inputs.append([])
                outputs.append([])
            features = self._create_input_array(inputs)
            labels = self._create_output_array(outputs)
            gather_indices = np.array([(i, len(x)) for i, x in enumerate(inputs)
                                      ])
            yield ([features, gather_indices,
                    np.array(self.get_global_step())], [labels], [])


class AspuruGuzikAutoEncoder(SeqToSeq):
    """
    This is an implementation of Automatic Chemical Design Using a Continuous Representation of Molecules
    http://pubs.acs.org/doi/full/10.1021/acscentsci.7b00572

    Abstract
    --------
    We report a method to convert discrete representations of molecules to and
    from a multidimensional continuous representation. This model allows us to
    generate new molecules for efficient exploration and optimization through
    open-ended spaces of chemical compounds. A deep neural network was trained on
    hundreds of thousands of existing chemical structures to construct three
    coupled functions: an encoder, a decoder, and a predictor. The encoder
    converts the discrete representation of a molecule into a real-valued
    continuous vector, and the decoder converts these continuous vectors back to
    discrete molecular representations. The predictor estimates chemical
    properties from the latent continuous vector representation of the molecule.
    Continuous representations of molecules allow us to automatically generate
    novel chemical structures by performing simple operations in the latent space,
    such as decoding random vectors, perturbing known chemical structures, or
    interpolating between molecules. Continuous representations also allow the use
    of powerful gradient-based optimization to efficiently guide the search for
    optimized functional compounds. We demonstrate our method in the domain of
    drug-like molecules and also in a set of molecules with fewer that nine heavy
    atoms.

    Notes
    -------
    This is currently an imperfect reproduction of the paper.  One difference is
    that teacher forcing in the decoder is not implemented.  The paper also
    discusses co-learning molecular properties at the same time as training the
    encoder/decoder.  This is not done here.  The hyperparameters chosen are from
    ZINC dataset.

    This network also currently suffers from exploding gradients.  Care has to be taken when training.

    NOTE(LESWING): Will need to play around with annealing schedule to not have exploding gradients
    TODO(LESWING): Teacher Forcing
    TODO(LESWING): Sigmoid variational loss annealing schedule
    The output GRU layer had one
    additional input, corresponding to the character sampled from the softmax output of the
    previous time step and was trained using teacher forcing. 48 This increased the accuracy
    of generated SMILES strings, which resulted in higher fractions of valid SMILES strings
    for latent points outside the training data, but also made training more difficult, since the
    decoder showed a tendency to ignore the (variational) encoding and rely solely on the input
    sequence. The variational loss was annealed according to sigmoid schedule after 29 epochs,
    running for a total 120 epochs

    I also added a BatchNorm before the mean and std embedding layers.  This has empiracally
    made training more stable, and is discussed in Ladder Variational Autoencoders.
    https://arxiv.org/pdf/1602.02282.pdf
    Maybe if Teacher Forcing and Sigmoid variational loss annealing schedule are used the
    BatchNorm will no longer be neccessary.
    """

    def __init__(self,
                 num_tokens,
                 max_output_length,
                 embedding_dimension=196,
                 filter_sizes=[9, 9, 10],
                 kernel_sizes=[9, 9, 11],
                 decoder_dimension=488,
                 **kwargs):
        """
        Parameters
        ----------
        filter_sizes: list of int
            Number of filters for each 1D convolution in the encoder
        kernel_sizes: list of int
            Kernel size for each 1D convolution in the encoder
        decoder_dimension: int
            Number of channels for the GRU Decoder
        """
        if len(filter_sizes) != len(kernel_sizes):
            raise ValueError("Must have same number of layers and kernels")
        self._filter_sizes = filter_sizes
        self._kernel_sizes = kernel_sizes
        self._decoder_dimension = decoder_dimension
        super(AspuruGuzikAutoEncoder,
              self).__init__(input_tokens=num_tokens,
                             output_tokens=num_tokens,
                             max_output_length=max_output_length,
                             embedding_dimension=embedding_dimension,
                             variational=True,
                             reverse_input=False,
                             **kwargs)

    def _create_features(self):
        return Input(shape=(self._max_output_length, len(self._input_tokens)))

    def _create_encoder(self, n_layers, dropout):
        """Create the encoder as a tf.keras.Model."""
        input = self._create_features()
        gather_indices = Input(shape=(2,), dtype=tf.int32)
        prev_layer = input
        for i in range(len(self._filter_sizes)):
            filter_size = self._filter_sizes[i]
            kernel_size = self._kernel_sizes[i]
            if dropout > 0.0:
                prev_layer = Dropout(rate=dropout)(prev_layer)
            prev_layer = Conv1D(filters=filter_size,
                                kernel_size=kernel_size,
                                activation=tf.nn.relu)(prev_layer)
        prev_layer = Flatten()(prev_layer)
        prev_layer = Dense(self._decoder_dimension,
                           activation=tf.nn.relu)(prev_layer)
        prev_layer = BatchNormalization()(prev_layer)
        return tf.keras.Model(inputs=[input, gather_indices],
                              outputs=prev_layer)

    def _create_decoder(self, n_layers, dropout):
        """Create the decoder as a tf.keras.Model."""
        input = Input(shape=(self._embedding_dimension,))
        prev_layer = Dense(self._embedding_dimension,
                           activation=tf.nn.relu)(input)
        prev_layer = layers.Stack()(self._max_output_length * [prev_layer])
        for i in range(3):
            if dropout > 0.0:
                prev_layer = Dropout(dropout)(prev_layer)
            prev_layer = GRU(self._decoder_dimension,
                             return_sequences=True)(prev_layer)
        output = Dense(len(self._output_tokens),
                       activation=tf.nn.softmax)(prev_layer)
        return tf.keras.Model(inputs=input, outputs=output)

    def _create_input_array(self, sequences):
        return self._create_output_array(sequences)



"""
Contains an abstract base class that supports different ML models.
"""

import os
import shutil
import tempfile
import logging
from typing import List, Optional, Sequence

import numpy as np

from deepchem.data import Dataset
from deepchem.metrics import Metric
from deepchem.trans import Transformer, undo_transforms
from deepchem.utils.evaluate import Evaluator
from deepchem.utils.typing import OneOrMany

logger = logging.getLogger(__name__)


class Model(object):
    """
    Abstract base class for DeepChem models.
    """

    def __init__(self,
                 model=None,
                 model_dir: Optional[str] = None,
                 **kwargs) -> None:
        """Abstract class for all models.

        This is intended only for convenience of subclass implementations
        and should not be invoked directly.

        Parameters
        ----------
        model: object
            Wrapper around ScikitLearn/Keras/Tensorflow model object.
        model_dir: str, optional (default None)
            Path to directory where model will be stored. If not specified,
            model will be stored in a temporary directory.
        """
        if self.__class__.__name__ == "Model":
            raise ValueError(
                "This constructor is for an abstract class and should never be called directly."
                "Can only call from subclass constructors.")

        self.model_dir_is_temp = False
        if model_dir is not None:
            if not os.path.exists(model_dir):
                os.makedirs(model_dir)
        else:
            model_dir = tempfile.mkdtemp()
            self.model_dir_is_temp = True
        self.model_dir = model_dir
        self.model = model
        self.model_class = model.__class__

    def __del__(self):
        if 'model_dir_is_temp' in dir(self) and self.model_dir_is_temp:
            shutil.rmtree(self.model_dir)

    def fit_on_batch(self, X: Sequence, y: Sequence, w: Sequence):
        """Perform a single step of training.

        Parameters
        ----------
        X: np.ndarray
            the inputs for the batch
        y: np.ndarray
            the labels for the batch
        w: np.ndarray
            the weights for the batch
        """
        raise NotImplementedError(
            "Each model is responsible for its own fit_on_batch method.")

    def predict_on_batch(self, X: np.typing.ArrayLike):
        """
        Makes predictions on given batch of new data.

        Parameters
        ----------
        X: np.ndarray
            Features
        """
        raise NotImplementedError(
            "Each model is responsible for its own predict_on_batch method.")

    def reload(self) -> None:
        """
        Reload trained model from disk.
        """
        raise NotImplementedError(
            "Each model is responsible for its own reload method.")

    @staticmethod
    def get_model_filename(model_dir: str) -> str:
        """
        Given model directory, obtain filename for the model itself.
        """
        return os.path.join(model_dir, "model.joblib")

    @staticmethod
    def get_params_filename(model_dir: str) -> str:
        """
        Given model directory, obtain filename for the model itself.
        """
        return os.path.join(model_dir, "model_params.joblib")

    def save(self) -> None:
        """Dispatcher function for saving.

        Each subclass is responsible for overriding this method.
        """
        raise NotImplementedError

    def fit(self, dataset: Dataset):
        """
        Fits a model on data in a Dataset object.

        Parameters
        ----------
        dataset: Dataset
            the Dataset to train on
        """
        raise NotImplementedError(
            "Each model is responsible for its own fit method.")

    def predict(self,
                dataset: Dataset,
                transformers: List[Transformer] = []) -> OneOrMany[np.ndarray]:
        """
        Uses self to make predictions on provided Dataset object.

        Parameters
        ----------
        dataset: Dataset
            Dataset to make prediction on
        transformers: List[Transformer]
            Transformers that the input data has been transformed by. The output
            is passed through these transformers to undo the transformations.

        Returns
        -------
        np.ndarray
            A numpy array of predictions the model produces.
        """
        y_preds = []
        for (X_batch, _, _,
             ids_batch) in dataset.iterbatches(deterministic=True):
            n_samples = len(X_batch)
            y_pred_batch = self.predict_on_batch(X_batch)
            # Discard any padded predictions
            y_pred_batch = y_pred_batch[:n_samples]
            y_pred_batch = undo_transforms(y_pred_batch, transformers)
            y_preds.append(y_pred_batch)
        y_pred = np.concatenate(y_preds)
        return y_pred

    def evaluate(self,
                 dataset: Dataset,
                 metrics: List[Metric],
                 transformers: List[Transformer] = [],
                 per_task_metrics: bool = False,
                 use_sample_weights: bool = False,
                 n_classes: int = 2):
        """
        Evaluates the performance of this model on specified dataset.

        This function uses `Evaluator` under the hood to perform model
        evaluation. As a result, it inherits the same limitations of
        `Evaluator`. Namely, that only regression and classification
        models can be evaluated in this fashion. For generator models, you
        will need to overwrite this method to perform a custom evaluation.

        Keyword arguments specified here will be passed to
        `Evaluator.compute_model_performance`.

        Parameters
        ----------
        dataset: Dataset
            Dataset object.
        metrics: Metric / List[Metric] / function
            The set of metrics provided. This class attempts to do some
            intelligent handling of input. If a single `dc.metrics.Metric`
            object is provided or a list is provided, it will evaluate
            `self.model` on these metrics. If a function is provided, it is
            assumed to be a metric function that this method will attempt to
            wrap in a `dc.metrics.Metric` object. A metric function must
            accept two arguments, `y_true, y_pred` both of which are
            `np.ndarray` objects and return a floating point score. The
            metric function may also accept a keyword argument
            `sample_weight` to account for per-sample weights.
        transformers: List[Transformer]
            List of `dc.trans.Transformer` objects. These transformations
            must have been applied to `dataset` previously. The dataset will
            be untransformed for metric evaluation.
        per_task_metrics: bool, optional (default False)
            If true, return computed metric for each task on multitask dataset.
        use_sample_weights: bool, optional (default False)
            If set, use per-sample weights `w`.
        n_classes: int, optional (default None)
            If specified, will use `n_classes` as the number of unique classes
            in `self.dataset`. Note that this argument will be ignored for
            regression metrics.

        Returns
        -------
        multitask_scores: dict
            Dictionary mapping names of metrics to metric scores.
        all_task_scores: dict, optional
            If `per_task_metrics == True` is passed as a keyword argument,
            then returns a second dictionary of scores for each task
            separately.
        """
        evaluator = Evaluator(self, dataset, transformers)
        return evaluator.compute_model_performance(
            metrics,
            per_task_metrics=per_task_metrics,
            use_sample_weights=use_sample_weights,
            n_classes=n_classes)

    def get_task_type(self) -> str:
        """
        Currently models can only be classifiers or regressors.
        """
        raise NotImplementedError

    def get_num_tasks(self) -> int:
        """
        Get number of tasks.
        """
        raise NotImplementedError
